
from epyk.core.css.styles.classes import CssStyle

from epyk.core.css import Defaults as Defaults_css


class CssDivNoMargin(CssStyle.Style):
  _attrs = {'margin': 0, 'padding': 0}


class CssDivNoBorder(CssStyle.Style):
  _attrs = {'margin': 0, 'clear': 'both', 'padding': 0, 'border': 0}


class CssDivBottomBorder(CssStyle.Style):
  def customize(self):
    self.css({"border-bottom": '2px solid %s' % self.rptObj.theme.colors[0], 'font-family': Defaults_css.Font.family})
    self.hover.css({"border-bottom": '2px solid %s' % self.rptObj.theme.success[-1]})


class CssDivWithBorder(CssStyle.Style):
  _attrs = {'margin': '0 0 5px 0', 'padding': '5px', 'outline': 'none'}

  def customize(self):
    self.css({'border': "1px solid %s" % self.rptObj.theme.colors[0], 'font-family': Defaults_css.Font.family})


class CssDivConsole(CssStyle.Style):
  _attrs = {'margin': 0, 'padding': '5px', 'border': 0, 'outline': 'none'}
  
  def customize(self):
    self.css({'background-color': self.rptObj.theme.greys[9]})


class CssDivCursor(CssStyle.Style):
  _attrs = {'cursor': 'pointer'}


class CsssDivBoxMarginVertical(CssStyle.Style):
  _attrs = {'margin': "5px 0"}


class CsssDivBoxMarginBorder(CssStyle.Style):
  _attrs = {'margin': 0, 'padding': '0 2px 0 2px', 'white-space': 'pre-wrap'}

  def customize(self):
    self.css({"border": '1px solid %s' % self.rptObj.theme.greys[0], 'font-family': Defaults_css.Font.family})
    self.hover.css({'border': "1px solid %s" % self.rptObj.theme.greys[5]})


class CssDivBoxCenter(CssStyle.Style):
  _attrs = {'width': '100%', 'text-align': 'center'}


class CssDivBoxWithDotBorder(CssStyle.Style):
  _attrs = {'margin': '5px'}

  def customize(self):
    self.css({"border": '1px dashed %s' % self.rptObj.theme.greys[9]})


class CssDivBubble(CssStyle.Style):
  _attrs = {'margin-left': 'auto', 'margin-right': 'auto', 'border-radius': '50%',
            'text-align': 'center'}

  def customize(self):
    self.css({"border": '1px solid %s' % self.rptObj.theme.greys[5]})


# class CssDivBox(CssStyle.Style):
#   _attrs = {'width': '100%'}
#
#   def customize(self):
#     self.hover.css({"border-left": "4px solid %s" % self.rptObj.theme.greys[0],
#                     'background-color': self.rptObj.theme.colors[9]})


class CssDivLeft(CssStyle.Style):
  _attrs = {'float': 'left', 'width': '20%'}


class CssDivRight(CssStyle.Style):
  _attrs = {'float': 'right', 'width': '80%'}


class CssDivBorder(CssStyle.Style):

  def customize(self):
    self.hover.css({"border": "1px solid %s" % self.rptObj.theme.greys[9]})


class CssDivShadow(CssStyle.Style):
  _attrs = {'box-shadow': '10px 10px 8px 10px #888888'}


# class CssDivWhitePage(CssStyle.Style):
#   _attrs = {'height': '80%', 'min-height': '600px', 'margin': '0 10px 0 10px',
#             'box-shadow': '10px 10px 8px 10px #888888'}
#
#   def customize(self):
#     self.hover.css({'background-color': self.rptObj.theme.greys[0],
#                     'border': '1px solid %s' % self.rptObj.theme.greys[9]})


class CssDivBanner(CssStyle.Style):
  _attrs = {'width': '100%', 'margin': 0, 'overflow-y': 'auto', 'padding': '10px'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.greys[0]})


class CssDivSubBanner(CssStyle.Style):
  _attrs = {'height': '400px', 'width': '100%', 'overflow-y': 'auto', 'margin-top': '50px', 'padding': 0, 'margin': 0}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[9], 'background-color': self.rptObj.theme.greys[0]})


class CssDivLabelPoint(CssStyle.Style):
  _attrs = {'padding': '10px', 'margin-top': '20px', 'margin-left': '5px', 'border-radius': '50%', 'cursor': 'pointer',
            'display': 'inline-block'}
  cssId = {'child': 'label'}

  def customize(self):
    self.css({"border": '1px solid %s' % self.rptObj.theme.greys[4], 'background': self.rptObj.theme.greys[0]})
    self.hover.css({"border": '1px solid %s' % self.rptObj.theme.success[1]})


class CssDivCommBubble(CssStyle.Style):
  _attrs = {'width': '100%', 'vertical-align': 'top', 'top': 0, 'margin-bottom': '20px', 'margin-left': '20px',
            'display': 'inline-block'}
  _before = {'content': "''", 'width': 0, 'height': 0, 'display': 'inline-block', 'border': '15px solid transparent',
             'margin-left': '-30px'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[0]})
    self.before.css({'border-right-color': self.rptObj.theme.colors[9]})


class CssDivComms(CssStyle.Style):
  _attrs = {'margin-top': '10px', 'padding': '5px'}


class CssDivLoading(CssStyle.Style):
  _attrs = {'opacity': 0.5, 'filter': 'alpha(opacity=50)', 'padding': '5px', 'text-align': 'center'}


class CssDivHidden(CssStyle.Style):
  _attrs = {'display': 'none'}


class CssDivTextLeft(CssStyle.Style):
  _attrs = {'text-align': 'left'}


class CssDivTableContent(CssStyle.Style):
  _attrs = {'padding': '5px 10px 5px 10px', 'width': 'auto', 'display': 'inline-block'}

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.greys[3], 'background-color': self.rptObj.theme.greys[0]})


class CssDivPagination(CssStyle.Style):
  _attrs = {'margin': 'auto', 'padding': '8px 16px', 'text-decoration': 'none', 'transition': 'background-color .3s'}
  _selectors = {'child': 'a'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[9]})
    self.hover.css({"background-color": self.rptObj.theme.greys[3]})


class CssDivEditor(CssStyle.Style):
  _attrs = {'overflow': 'hidden', 'white-space': 'pre', 'display': 'block', 'padding': '30px 10px 10px 10px',
            'margin-top': '5px', 'text-align': 'left'}

  def customize(self):
    self.css({'border': "1px solid %s" % self.rptObj.theme.greys[3], 'background-color': self.rptObj.theme.greys[2]})
    self.focus.css({'background-color': self.rptObj.theme.greys[0], 'border': "2px solid %s" % self.rptObj.theme.colors[5]})


class CssDivRow(CssStyle.Style):

  def customize(self):
    self.hover.css({'background-color': self.rptObj.theme.greys[1]})


class CssPanelTitle(CssStyle.Style):
  _attrs = {'padding': '1px 0', 'margin': '0 5px 5px 5px', 'font-weight': 'bold'}

  def customize(self):
    self.css({'border-bottom': "1px solid %s" % self.rptObj.theme.success[1],
              'font-size': '%s%s' % (Defaults_css.Font.header_size, Defaults_css.Font.unit),
              'font-family': Defaults_css.Font.family})


class CssDivFilter(CssStyle.Style):
  _attrs = {"padding": "5px"}

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.colors[0]})


class CssDivFilterItems(CssStyle.Style):
  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.colors[2]})
    self.hover.css({'border': '1px solid %s' % self.rptObj.theme.success[1]})


class CssDivModal(CssStyle.Style):
  _attrs = {'z-index': 100, 'position': 'fixed', 'padding-top': '100px', 'left': 0, 'top': 0,
            'width': '100%', 'height': '100%', 'overflow': 'auto', 'text-align': 'center'}


class CssDivModalContent(CssStyle.Style):
  _attrs = {'margin': '10%', 'padding': '5px 5px 5px 5px', 'border': '1px solid #888', 'width': '75%',
            'box-shadow': '0 19px 38px rgba(0, 0, 0, 0.12), 0 15px 12px rgba(0, 0, 0, 0.22)',
            'display': 'inline-flex', 'flex-direction': 'column'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.greys[0]})
    self.animation('epyk_modal_animatetop', {'0%': {'top': '-300px', 'opacity': '0'}, '100%': {'top': '0px', 'opacity': '1'}}, 0.7, iteration=1)


class CssDivStepper(CssStyle.Style):
  _selectors = {'child': 'li'}

  _attrs = {'float': 'left', 'text-align': 'center'}



import time

from epyk.core.html import Html

from epyk.core.html.options import OptInputs


class DatePicker(Html.Html):
  __reqCss, __reqJs = ['jqueryui'], ['jqueryui']
  name, category, callFnc = 'Date Picker', 'Dates', 'date'

  def __init__(self, report, value, label, icon, color, htmlCode, profile, options, helper):
    dfltOptions = {'dateFormat': 'yy-mm-dd'}
    dfltOptions.update(options)
    super(DatePicker, self).__init__(report, value, htmlCode=htmlCode, profile=profile)
    # Add all the internal components input, label, icon and helper
    self.input = self._report.ui.inputs.d_date(self.val, options=dfltOptions).css({"padding": 0})
    self.prepend_child(self.input)
    self.add_icon(icon, css={"margin-left": '5px', 'color': self._report.theme.success[1]}, position="after")
    if self.icon is not None:
      self.icon.click(self.input.dom.events.trigger("click").toStr())
    self.add_label(label, css={"padding": '2px 0', 'height': 'auto'})
    self.add_helper(helper, css={"float": "none", "margin-left": "5px"})
    self.css({"color": color or 'inherit', "vertical-align": "middle", "display": "block", "width": 'auto',
              'margin-top': '2px'})

  @property
  def options(self):
    """
    Description:
    -----------
    The progress bar is designed to display the current percent complete for a process.
    The bar is coded to be flexibly sized through CSS and will scale to fit inside its parent container by default.

    Related Pages:

			https://api.jqueryui.com/menu

    :rtype: OptInputs.OptionsDatePicker
    """
    return self.input.options

  def excluded_dates(self, dts=None, jsFncs=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param dts:
    :param jsFncs:
    """
    return self.input.excluded_dates(dts, jsFncs)

  def included_dates(self, dts=None, jsFncs=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param dts:
    :param jsFncs:
    """
    return self.input.included_dates(dts, jsFncs)

  def add_options(self, options=None, name=None, value=None):
    """
    Add TimePicker options

    Related Pages:

			https://timepicker.co/options/

    Attributes:
    ----------
    :param key: A string or a Python dictionary with the options to set
    :param val: Optional.
    """
    if options is None and name is None:
      raise Exception("Either the attrs or the name should be specified")

    if options is None:
      options = {name: value}
    for k, v in options.items():
      self.vals['options'][k] = v
    return self

  def __str__(self):
    return '<div %(attr)s>%(helper)s</div>' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'helper': self.helper}


class TimePicker(Html.Html):
  __reqCss, __reqJs = ['timepicker'], ['timepicker']
  name, category, callFnc = 'Time Picker', 'Dates', 'date'

  def __init__(self, report, value, label, icon, color, htmlCode, profile, options, helper):
    super(TimePicker, self).__init__(report, None, htmlCode=htmlCode, profile=profile)
    # Add the internal components (label, icon)
    self.input = self._report.ui.inputs.d_time(value, options=options)
    self.input.set_attrs(name="class", value='time').css({"padding": 0})
    self.prepend_child(self.input)
    self.add_icon(icon, css={"margin-left": '5px', 'color': self._report.theme.success[1]}, position="after")
    if self.icon is not None:
      self.icon.click(self.input.dom.events.trigger("click").toStr())
    self.add_label(label, css={"padding": '2px 0', 'height': 'auto'})
    self.add_helper(helper, css={"float": "none", "margin-left": "5px"})
    self.css({"color": color or 'inherit', "vertical-align": "middle", 'margin-top': '2px'})

  @property
  def options(self):
    """
    Description:
    -----------
    The progress bar is designed to display the current percent complete for a process.
    The bar is coded to be flexibly sized through CSS and will scale to fit inside its parent container by default.

    Related Pages:

			https://api.jqueryui.com/menu

    :rtype: OptInputs.OptionsTimePicker
    """
    return self.input.options

  def change(self, jsFnc):
    """
    Description:
    -----------
    Event triggerd when the value of the input field changes. A Date object containing the selected time is passed as the first argument of the callback.
    Note: the variable time is a function parameter received in the Javascript side

    Usage::

      morning = rptObj.ui.fields.time("8:13:00", label="Time field")
    morning.change([
      rptObj.js.alert("time", skip_data_convert=True)
    ])

    Related Pages:

			https://timepicker.co/options/

    :param jsFnc:
    """
    self.input.change(jsFnc)
    return self

  def __str__(self):
    return '<div %(attr)s>%(helper)s</div>' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'helper': self.helper}


class CountDownDate(Html.Html):
  name, category, callFnc = 'Countdown', 'Dates', 'countdown'

  def __init__(self, report, yyyy_mm_dd, label, icon, timeInMilliSeconds, width, height, htmlCode, helper, profile):
    super(CountDownDate, self).__init__(report, yyyy_mm_dd, code=htmlCode, profile=profile,
                                        css_attrs={"width": width, "height": height})
    self._jsStyles = {"delete": True}
    self.timeInMilliSeconds = timeInMilliSeconds
    # Add the underlying components
    self.add_label(label, css={"padding": '2px 0', 'height': 'auto'})
    self.add_icon(icon)
    self.add_helper(helper)

  @property
  def jqId(self):
    return "$('#%s span')" % self.htmlId

  @property
  def _js__builder__(self):
    return ''' 
      var splitDt = data.split("-"); var endDate = new Date(splitDt[0], parseInt(splitDt[1])-1, splitDt[2]);
      var now = new Date().getTime(); var distance = endDate.getTime() - now;

      var days = Math.floor(distance / (1000 * 60 * 60 * 24));
      var hours = Math.floor((distance %% (1000 * 60 * 60 * 24)) / (1000 * 60 * 60));
      var minutes = Math.floor((distance %% (1000 * 60 * 60)) / (1000 * 60));
      var seconds = Math.floor((distance %% (1000 * 60)) / 1000);

      htmlObj.innerHTML = "<b>"+ days +"d "+ hours +"h "+ minutes + "m "+ seconds +"s </b>"; 
      if ((distance < 0) && (options.delete)){clearInterval(htmlObj.id +"_interval")
      }''' % {"report_name": self._report.run.report_name, "script_name": self._report.run.script_name}

  def __str__(self):
    self.jsUpdateDataFnc = '''var %(htmlId)s_interval = setInterval(function(){%(refresh)s}, %(timeInMilliSeconds)s)
              ''' % {'htmlId': self.htmlId, 'refresh': self.refresh(), 'timeInMilliSeconds': self.timeInMilliSeconds}
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.jsUpdateDataFnc)
    return '<div %s><span></span>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class LastUpdated(Html.Html):
  name, category, callFnc = 'Last Update', 'Text', 'update'

  def __init__(self, report, label, color, width, height, htmlCode, profile):
    super(LastUpdated, self).__init__(report, "%s %s" % (label or "Last update", time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())),
                                      htmlCode, css_attrs={"width": width, "height": height, "color": color}, profile=profile)

  def __str__(self):
    return '<div %(strAttr)s>%(content)s</div>' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'content': self.val}



"""
CSS Style module for the common Menu Bar components
"""

from epyk.core.css.styles.classes import CssStyle

from epyk.core.css import Defaults as Defaults_css


class CssSideBarMenu(CssStyle.Style):
  _attrs = {'display': 'block', 'width': '100%'}
  _hover = {'text-decoration': 'underline'}

  def customize(self):
    self.css({'color': self.rptObj.theme.colors[9]})


class CssSideBarFixed(CssStyle.Style):
  _attrs = {'height': '100%', 'width': '40px', 'text-align': 'center', 'position': 'fixed', 'padding': '60px 0 0 0',
            'z-index': 5, 'top': 0, 'left': 0, 'overflow-x': 'hidden'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[9]})


class CssSideBarBubble(CssStyle.Style):
  _attrs = {'position': 'fixed', 'display': 'none', 'overflow-x': 'hidden', 'text-align': 'left',
            'height': '100%', 'min-width': '200px', 'padding': '10px', 'color': 'black'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[9], 'font-family': Defaults_css.Font.family,
              'font-size': '%s%s' % (Defaults_css.Font.size, Defaults_css.Font.unit)})


class CssSideBar(CssStyle.Style):
  _attrs = {'height': '100%', 'position': 'fixed', 'z-index': 5, 'left': 0, 'overflow-x': 'hidden', 'padding-top': '15px'}


class CssSideBarLiHref(CssStyle.Style):
  _attrs = {'color': 'white', 'margin': 0, 'padding': 0}
  _selectors = {'child': 'li ul a'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[9]})


class CssSideBarLi(CssStyle.Style):
  _attrs = {'color': 'white', 'list-style-type': 'none'}
  _selectors = {'child': 'li'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[9]})


class CssParamsBar(CssStyle.Style):
  _attrs = {'vertical-align': 'middle', 'white-space': 'nowrap', 'overflow-x': 'auto', 'overflow-y': 'hidden',
            'padding': '2px 10px', 'z-index': 10, 'width': '100%', 'position': 'fixed', 'left': 0, 'margin': 0}

  def customize(self):
    self.css({'border-top': "1px solid %s" % self.rptObj.theme.greys[6], 'background-color': self.rptObj.theme.greys[0]})




from epyk.core.data import DataClass


class AutoFill(DataClass):

  def activate(self):
    self.columns = ":not(:first-child)"
    return self

  @property
  def alwaysAsk(self):
    """
    Description:
    -----------
    Always ask the end user if an action should be taken or not.

    Related Pages:

			https://datatables.net/reference/option/autoFill.alwaysAsk
    """
    return self._attrs["alwaysAsk"]

  @alwaysAsk.setter
  def alwaysAsk(self, val):
    self._attrs["alwaysAsk"] = val

  @property
  def columns(self):
    """
    Description:
    -----------
    Select the columns that can be auto filled.

    Related Pages:

			https://datatables.net/reference/option/autoFill.columns
    """
    return self._attrs["alwaysAsk"]

  @columns.setter
  def columns(self, val):
    self._attrs["columns"] = val

  @property
  def enable(self):
    """
    Description:
    -----------
    Initial enablement state of AutoFill.

    Related Pages:

			https://datatables.net/reference/option/autoFill.enable
    """
    return self._attrs["enable"]

  @enable.setter
  def enable(self, val):
    self._attrs["enable"] = val

  @property
  def focus(self):
    """
    Description:
    -----------
    Action that will cause the auto fill drag handle to appear in a cell.

    Related Pages:

			https://datatables.net/reference/option/autoFill.focus
    """
    return self._attrs["focus"]

  @focus.setter
  def focus(self, val):
    self._attrs["focus"] = val

  @property
  def horizontal(self):
    """
    Description:
    -----------
    Enable / disable user ability to horizontally drag and fill.

    Related Pages:

			https://datatables.net/reference/option/autoFill.horizontal
    """
    return self._attrs["horizontal"]

  @horizontal.setter
  def horizontal(self, val):
    self._attrs["horizontal"] = val

  @property
  def update(self):
    """
    Description:
    -----------
    Control automatic update of data when a fill drag is completed.

    Related Pages:

			https://datatables.net/reference/option/autoFill.update
    """
    return self._attrs["update"]

  @update.setter
  def update(self, val):
    self._attrs["update"] = val

  @property
  def vertical(self):
    """
    Description:
    -----------
    Enable / disable user ability to vertically drag and fill.

    Related Pages:

			https://datatables.net/reference/option/autoFill.vertical
    """
    return self._attrs["vertical"]

  @vertical.setter
  def vertical(self, val):
    self._attrs["vertical"] = val



"""
Wrapper for the Javascript Math module

Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_math.asp
"""


from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsNumber


class JsMaths(object):

  @property
  def E(self):
    """
    The E property returns the Euler's number and the base of natural logarithms, approximately 2.718.

    Example
    jsObj.math.E

    Related Pages:

			https//www.w3schools.com/jsref/jsref_e.asp

    :return: Returns Euler's number (approx. 2.718)
    """
    return JsNumber.JsNumber("Math.E", isPyData=False)

  @property
  def LN2(self):
    """
    The LN2 property returns the natural logarithm of 2, approximately 0.693.

    Example
    jsObj.math.LN2

    Related Pages:

			https//www.w3schools.com/jsref/jsref_ln2.asp

    :return: Returns the natural logarithm of 2 (approx. 0.693)
    """
    return JsNumber.JsNumber("Math.LN2", isPyData=False)

  @property
  def LN10(self):
    """
    The LN10 property returns the natural logarithm of 10, approximately 2.302.

    Example
    jsObj.math.LN10

    Related Pages:

			https://www.w3schools.com/jsref/jsref_ln10.asp

    :return: Returns the natural logarithm of 10 (approx. 2.302)
    """
    return JsNumber.JsNumber("Math.LN10", isPyData=False)

  @property
  def LOG2E(self):
    """
    The LOG2E property returns the base-2 logarithm of E, approximately 1.442

    Example
    jsObj.math.LOG2E

    Related Pages:

			https//www.w3schools.com/jsref/jsref_log2e.asp

    :return: Returns the base-2 logarithm of E (approx. 1.442)
    """
    return JsNumber.JsNumber("Math.LOG2E", isPyData=False)

  @property
  def SQRT1_2(self):
    """
    The SQRT1_2 property returns the square root of 1/2, approximately 0.707.

    Example
    jsObj.math.SQRT1_2

    Related Pages:

			https//www.w3schools.com/jsref/jsref_sqrt1_2.asp

    :return: Returns the square root of 1/2 (approx. 0.707)
    """
    return JsNumber.JsNumber("Math.SQRT1_2", isPyData=False)

  @property
  def SQRT2(self):
    """
    The SQRT2 property returns the square root of 2, approximately 1.414.

    Example
    jsObj.math.SQRT2

    Example
    rptObj.js.math.SQRT2

    Related Pages:

			https//www.w3schools.com/jsref/jsref_sqrt2.asp

    :return: Returns the square root of 2 (approx. 1.414)
    """
    return JsNumber.JsNumber("Math.SQRT2", isPyData=False)

  @property
  def PI(self):
    """
    The PI property returns the ratio of a circle's area to the square of its radius, approximately 3.14

    https://www.w3schools.com/jsref/jsref_pi.asp
    """
    return JsNumber.JsNumber("Math.PI", isPyData=False)

  def random(self, min=0, max=1):
    """
    Math.random() returns a random number between 0 (inclusive),  and 1 (exclusive):

    Example
    rptObj.js.math.random()
    jsObj.math.random(10, 100)

    Related Pages:

			https://www.w3schools.com/js/js_random.asp

    :return: A Number, representing a number from 0 up to but not including 1
    """
    if min == 0 and max == 1:
      return JsNumber.JsNumber("Math.random()", isPyData=False)

    min = JsUtils.jsConvertData(min, None)
    max = JsUtils.jsConvertData(max, None)
    return JsNumber.JsNumber("Math.random() * (%(max)s - %(min)s + 1) + %(min)s" % {"min": min, "max": max})

  def min(self, *args):
    """
    The min() method returns the number with the lowest value.

    Example
    jsObj.math.min(10, 45, 100, -3, 56)

    Related Pages:

			https://www.w3schools.com/jsref/jsref_min.asp

    :param args: Optional. One or more numbers to compare

    :return: A Number, representing the lowest number of the arguments, or Infinity if no arguments are given, or NaN if one or more arguments are not numbers
    """
    jsArgs = [JsUtils.jsConvertData(a, None) for a in args]
    return JsNumber.JsNumber("Math.min(%s)" % ",".join([str(jsa) for jsa in jsArgs]), isPyData=False)

  def max(self, *args):
    """
    The max() method returns the number with the highest value.

    Example
    jsObj.math.max(10, 45, 100, -3, 56)

    Related Pages:

			https://www.w3schools.com/jsref/jsref_max.asp
    https://www.jstips.co/en/javascript/calculate-the-max-min-value-from-an-array/

    :param args: Optional. One or more numbers to compare

    :return: A Number, representing the highest number of the arguments, or -Infinity if no arguments are given, or NaN if one or more arguments are not numbers
    """
    jsArgs = [JsUtils.jsConvertData(a, None) for a in args]
    if len(jsArgs) == 1 and getattr(jsArgs[0], '_jsClass', None) == "Array":
      # ES2015 use of the new spread operator
      jsArgs[0] = "...%s" % jsArgs[0]
    return JsNumber.JsNumber("Math.max(%s)" % ",".join([str(jsa) for jsa in jsArgs]), isPyData=False)

  def floor(self, number):
    """
    The floor() method rounds a number DOWNWARDS to the nearest integer, and returns the result.

    Example
    jsObj.math.floor(13.566)

    Related Pages:

			https//www.w3schools.com/jsref/jsref_floor.asp

    :param number: Required. The number you want to round

    :return: A Number, representing the nearest integer when rounding downwards
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.floor(%s)" % number, isPyData=False)

  def trunc(self, number):
    """
    The trunc() method returns the integer part of a number.

    Example
    rptObj.js.math.trunc(rptObj.js.math.SQRT2)

    Related Pages:

			https//www.w3schools.com/jsref/jsref_trunc.asp

    :param number: Required. A number

    :return: Returns the integer part of a number (x)
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.trunc(%s)" % number, isPyData=False)

  def abs(self, number):
    """
    The abs() method returns the absolute value of a number.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_abs.asp

    :param number: Required. A number

    :return: Returns the absolute value of x
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.abs(%s)" % number, isPyData=False)

  def cos(self, number):
    """
    The acos() method returns the cosinus of a number as a value value between 0 and PI radians.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_cos.asp

    :param number: Returns the cosine of x (x is in radians)

    :return: A Number, from -1 to 1, representing the cosine of an angle, or NaN if the value is empty
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.cos(%s)" % number, isPyData=False)

  def sin(self, number):
    """
    The sin() method returns the sinus of a number as a value value between 0 and PI radians.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_sin.asp

    :param number: Returns the sinus of x (x is in radians)

    :return: A Number, from -1 to 1, representing the sine of an angle, or NaN if the value is empty
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.sin(%s)" % number, isPyData=False)

  def log(self, number):
    """
    The log() method returns the natural logarithm (base E) of a number.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_log.asp

    :param number: Required. A number

    :return: Returns the natural logarithm (base E) of x
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.log(%s)" % number, isPyData=False)

  def exp(self, number):
    """
    The exp() method returns the value of Ex, where E is Euler's number (approximately 2.7183) and x is the number passed to it.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_exp.asp

    :param number: Required. A number

    :return: Returns the value of exponential of x
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.exp(%s)" % number, isPyData=False)

  def round(self, number):
    """
    The round() method rounds a number to the nearest integer.

    Note: 2.49 will be rounded down (2), and 2.5 will be rounded up (3).

    Example
    jsObj.objects.number.new(23.6, varName="MyNumber")
    jsObj.math.round(jsObj.objects.number.get("MyNumber"))

    Related Pages:

			https//www.w3schools.com/jsref/jsref_round.asp

    :param number: Required. The number to be rounded

    :return: Rounds x to the nearest integer
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.round(%s)" % number, isPyData=False)

  def sqrt(self, number):
    """
    The sqrt() method returns the square root of a number.

    Example
    jsObj.objects.number.new(23.6, varName="MyNumber")
    jsObj.math.sqrt(jsObj.objects.number.get("MyNumber"))

    Related Pages:

			https//www.w3schools.com/jsref/jsref_sqrt.asp

    :param number: Required. A number

    :return: A Number. If x is a negative number, NaN is returned
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.sqrt(%s)" % number, isPyData=False)

  def ceil(self, number):
    """
    The ceil() method rounds a number UPWARDS to the nearest integer, and returns the result.

    Example
    jsObj.math.ceil(jsObj.objects.number.get("MyNumber"))

    Related Pages:

			https//www.w3schools.com/jsref/jsref_ceil.asp

    :param number: Required. The number you want to round

    :return: Returns x, rounded upwards to the nearest integer
    """
    number = JsUtils.jsConvertData(number, None)
    return JsNumber.JsNumber("Math.ceil(%s)" % number, isPyData=False)

  @staticmethod
  def pow(number, power):
    """
    The pow() method returns the value of x to the power of y (xy).

    Example
    jsObj.objects.number.new(23.6, varName="MyNumber")
    jsObj.math.pow(jsObj.objects.number.get("MyNumber"), 2)

    Related Pages:

			https//www.w3schools.com/jsref/jsref_pow.asp

    :param number: Required. The base
    :param power: Required. The exponent

    :return: Returns the value of x to the power of y
    """
    number = JsUtils.jsConvertData(number, None)
    power = JsUtils.jsConvertData(power, None)
    return JsNumber.JsNumber("Math.pow(%s, %s)" % (number, power), isPyData=False)



"""
## Markdown Framework

This framework is following the Markdown standard in order to write static documentation.
As part of the creation of the different components some extended rules have been added in order to ensure the use of very bespoke components and also allow a great flexibility.
This documentation will show you the different markdowns rules already in place.

If you have any questions regarding the principle of Markdowns, please have a quick look at the [wikipage](https://fr.wikipedia.org/wiki/Markdown)
Also the below link could be quite interesting to get more examples:
[Markdown Cheatsheet](https://github.com/adam-p/markdown-here/wiki/Markdown-Cheatsheet)

Markdown syntax is a standard used in many programming language but also by website to make the first implementation much easier
"""

import re
import os
import json
import sys
import inspect
import importlib
import logging


class DocCollection(list):
  def __init__(self, report):
    self._report = report

  def breadcrumb(self, paths, options=None, lang="eng"):
    """

    :return:
    """
    bubbles = []
    rows = ["<a href='/api'>Framework</a> / <a href='/api?module=local'>Scripting API</a>"]
    if len(paths) > 1:
      rows[0] = "%s / ... <a href=''>%s</a>" % (rows[0], paths[0])
    if options is not None:
      for key, val in options.items():
        if paths[-1] == key:
          bubbles.append(
            "<div title='%s' style='cursor:pointer;padding:1px 5px;border:1px solid black;display:inline-block;margin:0 2px;border-radius:5px;color:white;background-color:%s'>%s</div>" % (val[lang], self._report.getColor("colors", 5), key))
        else:
          bubbles.append("<a href='/api?enum=%s' title='%s' style='cursor:pointer;padding:1px 5px;border:1px solid black;display:inline-block;margin:0 2px;border-radius:5px;color:black;text-decoration:none'>%s</a>" % (key, val[lang], key) )
      rows.append("<div style='margin:3px 0'>%s</div>" % "".join(bubbles))
    self.append("<div>%s</div>" % "".join(rows))

  def breadcrumb2(self, levels, select_path):
    """

    :param levels:
    :param select_path:

    :return:
    """
    rows = []
    for i, vals in enumerate(levels):
      bubbles = []
      for val in vals:
        if val.lower() == select_path[i].lower():
          bubbles.append("<div style='cursor:pointer;padding:1px 5px;border:1px solid black;display:inline-block;color:white;background-color:%s;margin:0 2px;border-radius:5px'>%s</div>" % (self._report.getColor("greys", 9), val) )
        else:
          bubbles.append("<div style='cursor:pointer;padding:1px 5px;border:1px solid black;display:inline-block;margin:0 2px;border-radius:5px'>%s</div>" % val)
      rows.append("<div style='margin:3px 0'>%s</div>" % "".join(bubbles))
    self.append("<div>%s</div>" % "".join(rows))
    return self

  def title(self, val, level=1, css_pmts=None):
    """

    :param val:
    :param level:
    :param css_pmts:

    :return:
    """
    if val != '':
      self.append("%s %s" % ("".join(['#'] * level), val), css_pmts)
    return self

  def formula(self, val):
    """

    :param val:
    :return:
    """
    if val != '':
      self.append('$$ %s $$' % val)
    return self

  def link(self, val, url, icon='fas fa-chevron-circle-right', cssPmts=None):
    """

    :param val:
    :param url:
    :param icon:
    :param cssPmts:

    :return:
    """
    self.append("&nbsp;&nbsp;[!(%s) %s](%s)" % (icon, val, url), cssPmts)
    return self

  def hr(self):
    """

    :return:
    """
    self.append("***")
    return self

  def table(self, header, records, category='base', cssPmts=None, pmts=None):
    """

    :param header:
    :param records:
    :param category:
    :param cssPmts:
    :param pmts:

    :return:
    """
    data = ["|".join(header)]
    for rec in records:
      if isinstance(rec, dict):
        row = [rec[h] for h in header]
      else:
        row = rec
      data.append("|".join(row))
    if pmts is not None:
      data.append("--%s" % ";".join(["%s:%s" % (k, v) for k, v in pmts.items()]))
    if cssPmts is not None:
      data.append("@%s" % ";".join(["%s:'%s'" % (k, v)   for k, v in cssPmts.items()]))
    self.append("\n".join(["---Table:%s" % category] + data + ["---"]))
    return self

  def append(self, val, css_pmts=None):
    """

    :param val:
    :param css_pmts:

    :return:
    """
    if isinstance(val, list):
      val= "\n".join(val)
    if css_pmts is not None:
      val = "%s css%s" % (val, json.dumps(css_pmts).replace('"', ''))
    for v in val.split("\n\n"):
      super(DocCollection, self).append("%s\n" % v)
    return self

  def add(self, doc_obj, key):
    """

    :param doc_obj:
    :param key:

    :return:
    """
    self.append(doc_obj.getAttr(key))
    if doc_obj.source is not None:
      self.src(doc_obj.source)
    return self

  def src(self, source):
    """

    :param source:

    :return:
    """
    self.append("<div style='width:100%%;color:%s;font-style:italic;text-align:right;font-size:12px;padding-right:5px'>Source: %s</div>" % (self._report.getColor('greys', 4), source))
    return self

  def code(self, vals, language='python'):
    """

    :param vals:
    :param language:

    :return:
    """
    if isinstance(vals, list):
      vals= "\n".join(vals)
    if vals != '':
      match = re.search(">>>\((.*)\)", vals)
      if match:
        vals = vals.replace(match.group(0), '')
        vals = "%s\n>>> %s" % (vals, match.group(1))
      self.append("\n".join(["```%s" % language, vals.strip(), "```"]))
    return self


class DocString(dict):
  source = None

  def getAttr(self, dsc):
    return "\n".join(self.get(dsc, ['']))


class Convertor(object):
  def __init__(self, src=None):
    self.__src = src

  def icon(self, val):
    """

    :param val:
    :return:
    """
    pattern = re.compile("!\((.*)\)")
    res = pattern.search(val)
    if res is not None:
      for grp in res.groups():
        i = self.__src.ui.tags.i("")
        i.inReport = False
        i.attr["class"] = [grp]
        val = val.replace("!(%s)" % grp, i.html())
    return val

  def link(self, val):
    """

    :param val:
    :return:
    """
    pattern = re.compile("\[(.*)\]\((.*)\)")
    res = pattern.search(val)
    if res is not None:
      results = res.groups()
      for i in range(len(results)//2):
        text = results[2*i]
        url = results[2*i + 1]
        l = self.__src.ui.link(text, url)
        l.inReport = False
        val = val.replace("[%s](%s)" % (text, url), l.html())
    return val

  def bold(self, val):
    """
    Add the bold HTML tags to a string is ** markdown

    :param val: String. The text
    :return: String. The converted text
    """
    res = str(val).split("**")
    if len(res) > 1:
      tmp = []
      for i, r in enumerate(res):
        if i % 2 == 1:
          b = self.__src.ui.tags.b(r)
          b.inReport = False
          tmp.append(b.html())
        else:
          tmp.append(r)
      return "".join(tmp)

    return val

  def italic(self, val):
    """
    Add the Italic HTML tags to a string is * markdown

    :param val: String. The text
    :return: String. The converted text
    """
    res = val.split("*")
    if len(res) > 1:
      tmp = []
      for i, r in enumerate(res):
        if i % 2:
          i = self.__src.ui.tags.i(r)
          i.inReport = False
          tmp.append(i.html())
        else:
          tmp.append(r)
      return "".join(tmp)

    return val

  def highlighted(self, val):
    """
    Add the u HTML tags to a string is __ markdown

    :param val: String. The text
    :return: String. The converted text
    """
    res = str(val).split("__")
    if len(res) > 1:
      tmp = []
      for i, r in enumerate(res):
        if i % 2:
          u = self.__src.ui.tags.u(r)
          u.inReport = False
          tmp.append(u.html())
        else:
          tmp.append(r)
      return "".join(tmp)

    return val

  def deleted(self, val):
    """
    Add the deleted HTML tags to a string is -- markdown

    :param val: String. The text
    :return: String. The converted text
    """
    res = str(val).split("--")
    if len(res) > 1:
      tmp = []
      for i, r in enumerate(res):
        if i % 2:
          i = self.__src.ui.tags.delete(r)
          i.inReport = False
          tmp.append(i.html())
        else:
          tmp.append(r)
      return "".join(tmp)

    return val

  def all(self, val):
    """

    :param val:
    :return:
    """
    val = self.bold(str(val))
    val = self.highlighted(val)
    val = self.deleted(val)
    val = self.icon(val)
    val = self.link(val)
    val = self.italic(val)
    return val


class MarkDown(object):
  markDownMappings, markDownBlockMappings = None, None

  def __init__(self, full_path=None, report=None):
    self._report = report
    sys.path.append(os.path.join(os.getcwd(), 'epyk', 'core'))
    if full_path is not None:
      content = ""
      if full_path.endswith('amd'):
        py_file = open(full_path)
        content = py_file.read()
        py_file.close()
      self.content = content.split("\n\n")
    self.load_modules()

  def load_modules(self):
    """
    Fabric to load all the components in memory and use the match logic
    """
    # Store the modules
    from epyk.core import html
    from epyk.core.html import graph
    from epyk.core.html import tables

    self.markDownMappings, self.markDownBlockMappings = [], []
    for alias, mod in [('html', html), ('tables', tables), ('graph', graph)]:
      for script in os.listdir(os.path.dirname(mod.__file__)):
        if script.endswith(".py"):
          if alias in ['tables', 'graph']:
             mod_path = "epyk.core.html.%s.%s" % (alias, script.replace(".py", ""))
          else:
             mod_path = "epyk.core.html.%s" % script.replace(".py", "")
          for name, obj in inspect.getmembers(importlib.import_module(mod_path), inspect.isclass):
            try:
              if inspect.isclass(obj):
                if hasattr(obj, 'matchMarkDown'):
                  self.markDownMappings.append({'match': obj.matchMarkDown, 'convert': obj.convertMarkDown, 'name': name})
                elif hasattr(obj, 'matchMarkDownBlock'):
                  self.markDownBlockMappings.append({'match': obj.matchMarkDownBlock, 'convert': obj.convertMarkDownBlock, 'name': name, 'endBlock': obj.matchEndBlock})
            except Exception as err:
              logging.warning("%s, error %s" % (script, err))

  def find_block(self, block_split):
    """
    Function used to recognize block component

    :param block_split: The list of line which represents a component

    :return: The Ares Object
    """
    for blockMapping in self.markDownBlockMappings:
      if blockMapping['match'](block_split) is not None:
        return blockMapping

  def find_line(self, line, report):
    """

    :param line:
    :param report:

    :return:
    """
    for line_mapping in self.markDownMappings:
      reg_exp_result = line_mapping['match'](line)
      if reg_exp_result:
        return line_mapping['convert'](line, reg_exp_result, report)

  def convert(self, out_file_path):
    """

    :param out_file_path:

    :return:
    """
    out_file = open(out_file_path, 'w')
    converted_data = self.parse(self.content)
    for line in converted_data:
      out_file.write("%s\n" % line)
    for block in self.content:
      lines = block.split("\n")
    out_file.close()

  def load(self, report, data):
    """
    Convert the markdown static string to Python object.

    :param report:
    :param data:

    :return:
    """
    htmlObjs, j, spare_lines = [], 0, []
    if not isinstance(data, list):
      data = data.strip().split("\n\n")
    for i, block in enumerate(data):
      if j > 0:
        j -= 1
        continue

      lines = block.strip().split("\n")
      convert_block = self.find_block(lines)
      if convert_block is None:
        j = 0
        convert_line = self.find_line(lines[0].strip(), report)
        if convert_line is None:
          htmlObjs.append(report.ui.texts.paragraph(block))
        else:
          htmlObjs.append(report.htmlItems[report.content[-1]])
      else:
        block_data = lines
        for j, block in enumerate(data[i+1:]):
          if convert_block['endBlock'](lines[-1].strip()):
            break

          block_data.append("")
          lines = block.strip().split("\n")
          block_data.extend(lines)

        for v in convert_block['convert'](block_data, report):
          block_data.append(v)
        htmlObjs.append(report.htmlItems[report.content[-1]])
    return htmlObjs

  def parse(self, data, report=None, title=''):
    """

    :param data:
    :param report:
    :param title:

    :return:
    """
    if data is None:
      return ""

    out_file, j = ["TITLE='%s'" % title, "def report(rptObj):"], 0
    if not isinstance(data, list):
      data = data.split("\n\n")
    for i, block in enumerate(data):
      if j > 0:
        j -= 1
        continue

      lines = block.strip().split("\n")
      convert_block = self.find_block(lines)
      if convert_block is None:
        j = 0
        convertLine = self.find_line(lines[0], report)
        if convertLine is None:
          out_file.append("  rptObj.ui.texts.paragraph(%s)" % (json.dumps(block)))
          if report is not None:
            report.ui.texts.paragraph(block)
        else:
          for v in convertLine:
            out_file.append("  %s" % v)
      else:
        block_data = lines
        for j, block in enumerate(data[i+1:]):
          if convert_block['endBlock'](lines[-1].strip()):
            break

          block_data.append("")
          lines = block.strip().split("\n")
          block_data.extend(lines)
        complex_component = convert_block['convert'](block_data, report)
        for v in complex_component:
          block_data.append(v)
        out_file.append("  %s" % "".join(complex_component))
    return out_file

  @classmethod
  def parse_restructured_text(cls, text, report=None, lang='eng'):
    """
    Parse a reStructured type Documentation string
    This function will both:
      - return the string corresponding to the internal Markdown definition
      - Add the components to the report object (if defined)

    Example
    MarkDown.parse_restructured_text(MarkDown.loadsDsc.__doc__)

    Related Pages:

			https://en.wikipedia.org/wiki/ReStructuredText
    https://mail.python.org/pipermail/python-dev/2002-April/022131.html

    :param report: The report object
    :param text: The doc string to be parsed
    :param lang: The code for the language

    :return: A python dictionary with the documentation
    """
    doc_keywords = [("Documentation", 'documentation'), ("Example", 'examples')]
    ide_keywords = {'returns': re.compile(":return: (.*)"), 'params': re.compile(":param (.*): (.*)"),
                    'types': re.compile(":type (.*): (.*)")}
    section = 'desc'
    result = {'examples': [], 'documentation': [], 'desc': []}
    if text is None:
      return result

    for l in text.strip().split("\n"):
      line, is_section = l.strip(), False
      for doc_keyword, doc_tag in doc_keywords:
        if line == doc_keyword:
          section, is_section = doc_tag, True

      if not is_section:
        for reg_type, reg_keyword in ide_keywords.items():
          groups = reg_keyword.match(line)
          if groups is not None:
            if len(groups.groups()) > 1:
              result.setdefault(reg_type, {})[groups.group(1)] = groups.group(2)
            else:
              result.setdefault(reg_type, []).append(groups.group(1))
            break

        else:
          result.setdefault(section, []).append(line)
    return result

  def parse_module(self, module, report=None, lang='eng'):
    """
    Parse the documentation in the module

    Example
    md = PyMarkdown.MarkDown(report=Ares.Report()).parse_module(Js)

    :param module: An imported Python module
    :param report: The report object
    :param lang: The code for the language

    :return: A python dictionary with the module documentation
    """
    if report is None:
      report = self._report
    if report is not None:
      report.ui.title(module.__name__, level=1)
      self.parse(module.__doc__, report=report)
    result = {"classes": {}, "functions": {}, "content": module.__doc__}
    for name, obj in inspect.getmembers(module):
      if inspect.isfunction(obj):
        if report is not None:
          report.ui.title(name, level=3)

        fncDoc = self.parse_restructured_text(obj.__doc__, report, lang)
        self.parse(fncDoc['desc'], report=report)
        if len(fncDoc['examples']) > 0:
          report.ui.title("Examples", level=4)
          self.parse(fncDoc['examples'], report=report)
        result["functions"][name] = fncDoc
      elif inspect.isclass(obj):
        if obj.__module__ == module.__name__:
          if report is not None:
            report.ui.title(name, level=2)
          result["classes"][name] = self.parse_class(obj, report, lang)
    return result

  def parse_class(self, className, report=None, lang='eng'):
    """

    :param className:
    :param report:
    :param lang:

    :return: A python dictionary with the class documentation
    """
    result = {"properties": {}, "functions": {}}
    for name, obj in inspect.getmembers(className):
      if report is not None:
        report.ui.title(name, level=2)
      if inspect.isfunction(obj) and obj.__doc__ is not None:
        fnc_doc = self.parse_restructured_text(obj.__doc__, report, lang)
        self.parse(fnc_doc['desc'], report=report)
        if len(fnc_doc['examples']) > 0:
          report.ui.title("Examples", level=4)
          self.parse(fnc_doc['examples'], report=report)
        result["functions"][name] = fnc_doc
        #result["functions"][name] = cls.parse_restructured_text(obj.__doc__, report, lang)
      elif inspect.isdatadescriptor(obj) and obj.__doc__ is not None:
        result["properties"][name] = self.parse_restructured_text(obj.__doc__, report, lang)
    return result

  @classmethod
  def loads_dsc(cls, module_name, lang='eng'):
    """
    Load the documentation from the DSC variable of a given module.

    :param module_name: The Python Module Name
    :param lang: The code for the language

    :return: The docString object
    """
    if hasattr(module_name, 'DSC'):
      docStrObj = cls.loads(module_name.DSC.get(lang, module_name.DSC['eng']))
    else:
      docStrObj = cls.loads("")
    docStrObj.source = "/%s.py" % module_name.__name__.replace(".", "/")
    return docStrObj

  @staticmethod
  def loads(doc_string, section=None):
    """
    Load the documentation from a bespoke String variable.
    The string variable should have the expected labels in order to be correctly loaded and then parsed as a Markdown string

    :return: The docString object

    :param doc_string:
    :param section:
    :return:
    """
    if doc_string is None:
      return DocString()

    doc_string = doc_string.replace("https:", "https\\:").replace("http:", "http\\:")
    data, category = {}, None
    for com in doc_string.split("\n"):
      line = com.strip()
      if line.startswith(":"):
        match = re.search(":([a-zA-Z 0-9]*):(.*)", line)
        if match is not None:
          results = match.groups()
          category = results[0]
          if not category in data:
            data[category] = []
          doc_line = results[1].strip()
      else:
        doc_line = line
      if category is not None:
        data[category].append(doc_line)
    if section is not None:
      return DocString(data).getAttr(section)

    return DocString(data)





from epyk.core.js.fncs import JsFncs
from epyk.core.css import Colors

from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsString
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsBoolean
from epyk.core.js.primitives import JsArray

from epyk.core.js.objects import JsNodeDomRect

from epyk.core.js import JsUtils


class JsDomEvents(object):
  class __internal(object):
    htmlId = None

  def __init__(self, src=None):
    self._src = src if src is not None else self.__internal()
    self._js = []

  def stopPropagation(self):
    """
    Description:
    ------------
    The stopPropagation() method prevents propagation of the same event from being called.

    Related Pages:

			https://www.w3schools.com/jsref/event_stoppropagation.asp

    :return: The Python Dom object
    """
    self._js.append('stopPropagation()')
    return self

  def blur(self, jsFncs):
    """
    Description:
    ------------
    FocusEvent

    The event occurs when an element loses focus

    Related Pages:

			https://www.w3schools.com/jsref/event_onblur.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("blur", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def click(self, jsFncs):
    """
    Description:
    ------------
    The event occurs when the user clicks on an element

    Example
    select.label.dom.events.click(rptObj.js.console.log("test"))

    Related Pages:

			https://www.w3schools.com/jsref/event_onclick.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("click", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def change(self, jsFncs):
    """
    Description:
    ------------
    The event occurs when the content of a form element, the selection, or the checked state have changed (for <input>, <select>, and <textarea>)

    Example
    select.dom.events.change(rptObj.js.window.alert("test"))

    Related Pages:

			https://www.w3schools.com/jsref/event_onchange.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("change", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def dblclick(self, jsFncs):
    """
    Description:
    ------------
    The event occurs when the user double-clicks on an element

    Related Pages:

			https://www.w3schools.com/jsref/event_ondblclick.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("dblclick", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def focus(self, jsFncs):
    """
    Description:
    ------------
    The event occurs when an element gets focus

    Related Pages:

			https://www.w3schools.com/jsref/event_onfocusin.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("focus", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def focusin(self, jsFncs):
    """
    Description:
    ------------
    The event occurs when an element is about to get focus

    Related Pages:

			https://www.w3schools.com/jsref/event_onfocusin.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("focusin", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def focusout(self, jsFncs):
    """
    The event occurs when an element is about to lose focus

    Related Pages:

			https://www.w3schools.com/jsref/event_onfocusout.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("focusin", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def keydown(self, jsFncs):
    """
    The event occurs when the user is pressing a key

    Related Pages:

			https://www.w3schools.com/jsref/event_onkeydown.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("keydown", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def keypress(self, jsFncs):
    """
    The event occurs when the user presses a key

    Related Pages:

			https://www.w3schools.com/jsref/event_onkeypress.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("keypress", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def keyup(self, jsFncs):
    """
    The event occurs when the user releases a key

    Related Pages:

			https://www.w3schools.com/jsref/event_onkeyup.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("keyup", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def mousedown(self, jsFncs):
    """
    The event occurs when the user presses a mouse button over an element

    Related Pages:

			https://www.w3schools.com/jsref/event_onmousedown.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("mousedown", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def mouseenter(self, jsFncs):
    """
    The event occurs when the pointer is moved onto an element

    Related Pages:

			https://www.w3schools.com/jsref/event_onmouseenter.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("mousedown", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def mouseleave(self, jsFncs):
    """
    The event occurs when the pointer is moved out of an element

    Example
    select.label.dom.events.mouseleave(rptObj.js.console.log("test"))

    Related Pages:

			https://www.w3schools.com/jsref/event_onmouseleave.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("mouseleave", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def mouseover(self, jsFncs):
    """
    The event occurs when the pointer is moved onto an element, or onto one of its children

    Related Pages:

			https://www.w3schools.com/jsref/event_onmouseover.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("mouseover", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def mouseup(self, jsFncs):
    """
    The event occurs when a user releases a mouse button over an element

    Related Pages:

			https://www.w3schools.com/jsref/event_onmouseup.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("mouseover", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def mouseout(self, jsFncs):
    """
    The event occurs when a user releases a mouse button over an element

    Related Pages:

			https://www.w3schools.com/jsref/event_onmouseout.asp

    :param jsFncs: An array of Js functions or string. Or a string with the Js

    :return: The Python Dom object
    """
    self._js.append('addEventListener("mouseout", function(){%s})' % ";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def trigger(self, event, withFocus=True):
    """
    Trigger a javascript event

    Related Pages:

			https://www.w3schools.com/jsref/met_html_focus.asp
    https://developer.mozilla.org/en-US/docs/Web/Guide/Events/Creating_and_triggering_events

    :param event: The event name
    :param withFocus: Optional, a boolean to define if the focus needs to be set to this component

    :return: The Javascript string of this function
    """
    item = "document.getElementById('%(htmlId)s')" % {'htmlId': self._src.htmlId}
    if withFocus:
      return JsFncs.JsFunction('(function(){var clickEvent = new Event("%(event)s"); %(elem)s.focus(); %(elem)s.dispatchEvent(clickEvent)})()' % {"event": event, "elem": item})

    return JsFncs.JsFunction('(function(){var clickEvent = new Event("%(event)s"); %(elem)s.dispatchEvent(clickEvent)})()' % {"event": event, "elem": item})

  def toStr(self):
    if self._src.htmlId is None:
      raise Exception("Selector not defined, use this() or new() first")

    if len(self._js) == 0:
      return self._src.htmlId

    strData = "document.getElementById('%(htmlId)s').%(items)s" % {'htmlId': self._src.htmlId, 'items': ".".join(self._js)}
    self._js = [] # empty the stack
    return strData


class JsDomsTransforms(object):

  def __init__(self, src, selector):
    self._src, self.selector = src, selector

  def initial(self):
    """
    Sets this property to its default value

    """
    return "%s.initial" % self.selector

  def inherit(self):
    """
    Inherits this property from its parent element.

    """
    return "%s.initial" % self.selector

  def matrix(self, scaleX, skewY, skewX, scaleY, translateX, translateY):
    """
    Description:
    ------------
    Defines a 2D transformation, using a matrix of six values

    Usage::

      Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/matrix

    Attributes:
    ----------
    :param scaleX: Number.
    :param skewY: Number.
    :param skewX: Number.
    :param scaleY: Number.
    :param translateX: Number.
    :param translateY: Number.
    """
    return "%s.style.transform = 'matrix(%s, %s, %s, %s, %s, %s)'" % (self.selector, scaleX, skewY, skewX, scaleY, translateX, translateY)

  def translateX(self, x, unit='px'):
    """
    Description:
    ------------
    The translateX() CSS function repositions an element horizontally on the 2D plane. Its result is a <transform-function> data type.

    Usage::

      Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/translateX
    https://www.w3schools.com/jsref/prop_style_transform.asp

    Attributes:
    ----------
    :param x: Number.
    :param unit: String. The unit used for the transformation (px, cm, rem...)
    """
    return "%s.style.transform = 'translateX(%s%s)'" % (self.selector, x, unit)

  def translateY(self, y, unit='px'):
    """
    Description:
    ------------
    The translateY() CSS function repositions an element vertically on the 2D plane. Its result is a <transform-function> data type.

    Related Pages:

			https://www.w3schools.com/jsref/prop_style_transform.asp

    Attributes:
    ----------
    :param y:
    :param unit: String. The unit used for the transformation (px, cm, rem...)
    """
    return "%s.style.transform = 'translateY(%s%s)'" % (self.selector, y, unit)

  def translate(self, x, y, unit='px'):
    """
    Description:
    ------------
    The translate() CSS function repositions an element in the horizontal and/or vertical directions. Its result is a <transform-function> data type.

    Related Pages:

			https://www.w3schools.com/jsref/prop_style_transform.asp

    Attributes:
    ----------
    :param x:
    :param y:
    :param unit: String. The unit used for the transformation (px, cm, rem...)
    """
    return "%s.style.transform = 'translate(%s%s, %s%s)'" % (self.selector, x, unit, y, unit)

  def perspective(self, d, unit='px'):
    """
    Description:
    ------------
    The perspective() CSS function defines a transformation that sets the distance between the user and the z=0 plane,
    the perspective from which the viewer would be if the 2-dimensional interface were 3-dimensional.
    Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/perspective

    Attributes:
    ----------
    :param d: Is a <length> representing the distance from the user to the z=0 plane.
              If it is 0 or a negative value, no perspective transform is applied.
    :param unit: String. The unit used for the transformation (px, cm, rem...)
    """
    return "%s.style.transform = 'perspective(%s%s)'" % (self.selector, d, unit)

  def scale(self, x, y):
    """
    Description:
    ------------
    The scale() CSS function defines a transformation that resizes an element on the 2D plane.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/scale

    Attributes:
    ----------
    :param x: Number. A <number> representing the abscissa of the scaling vector.
    :param y: Number. A <number> representing the ordinate of the scaling vector.
              If not defined, its default value is sx, resulting in a uniform scaling that preserves the element's aspect ratio.
    """
    return "%s.style.transform = 'scale(%s, %s)'" % (self.selector, x, y)

  def scaleX(self, x):
    """
    Description:
    ------------
    The scaleX() CSS function defines a transformation that resizes an element along the x-axis (horizontally). Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/scaleX

    Attributes:
    ----------
    :param x: Number. Is a <number> representing the scaling factor to apply on the abscissa of each point of the element.
    """
    return "%s.style.transform = 'scaleX(%s)'" % (self.selector, x)

  def scaleY(self, y):
    """
    Description:
    ------------
    The scaleY() CSS function defines a transformation that resizes an element along the y-axis (vertically). Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/scaleY

    Attributes:
    ----------
    :param y: Number. Is a <number> representing the scaling factor to apply on the ordinate of each point of the element.
    """
    return "%s.style.transform = 'scaleY(%s)'" % (self.selector, y)

  def skew(self, angleX, angleY=0, unit='deg'):
    """
    Description:
    ------------
    The skew() CSS function defines a transformation that skews an element on the 2D plane. Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/skew

    Attributes:
    ----------
    :param angleX: Number. Is an <angle> representing the angle to use to distort the element along the abscissa.
    :param angleY: Number. Is an <angle> representing the angle to use to distort the element along the ordinate.
                           If not defined, its default value is 0, resulting in a purely horizontal skewing.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'skew(%s%s, %s%s)'" % (self.selector, angleX, unit, angleY, unit)

  def skewX(self, angle, unit='deg'):
    """
    Description:
    ------------
    The skewX() CSS function defines a transformation that skews an element in the horizontal direction on the 2D plane.
    Its result is a <transform-function> data type.

    Usage::

      i.label.dom.transform.skewX(20),

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/skewX

    Attributes:
    ----------
    :param angle: Number. Is an <angle> representing the angle to use to distort the element along the abscissa.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'skewX(%s%s)'" % (self.selector, angle, unit)

  def skewY(self, angle, unit='deg'):
    """
    Description:
    ------------
    The skewY() CSS function defines a transformation that skews an element in the vertical direction on the 2D plane.
    Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/skewY

    Attributes:
    ----------
    :param angle: Number. Is an <angle> representing the angle to use to distort the element along the ordinate.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'skewY(%s%s)'" % (self.selector, angle, unit)

  def rotate(self, r, unit='deg'):
    """
    Description:
    ------------
    The rotate() CSS function defines a transformation that rotates an element around a fixed point on the 2D plane, without deforming it.
    Its result is a <transform-function> data type.

    Usage::

      i.label.dom.transform.rotate(90)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/rotate

    Attributes:
    ----------
    :param r: Is an <angle> representing the angle of the rotation. A positive angle denotes a clockwise rotation, a negative angle a counter-clockwise one.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'rotate(%s%s)'" % (self.selector, r, unit)

  def rotate3d(self, x, y, z, a, unit='deg'):
    """
    Description:
    ------------
    The rotate3d() CSS function defines a transformation that rotates an element around a fixed axis in 3D space, without deforming it.
    Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/rotate3d

    Attributes:
    ----------
    :param x: Is a <number> describing the x-coordinate of the vector denoting the axis of rotation which could between 0 and 1.
    :param y: Is a <number> describing the y-coordinate of the vector denoting the axis of rotation which could between 0 and 1.
    :param z: Is a <number> describing the z-coordinate of the vector denoting the axis of rotation which could between 0 and 1.
    :param a: Is an <angle> representing the angle of the rotation. A positive angle denotes a clockwise rotation, a negative angle a counter-clockwise one.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'rotate3d(%s, %s, %s, %s%s)'" % (self.selector, x, y, z, a, unit)

  def rotateX(self, r, unit='deg'):
    """
    Description:
    ------------
    The rotateX() CSS function defines a transformation that rotates an element around the abscissa (horizontal axis) without deforming it.
    Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/rotateX

    Attributes:
    ----------
    :param r: Is an <angle> representing the angle of the rotation. A positive angle denotes a clockwise rotation, a negative angle a counter-clockwise one.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'rotateX(%s%s)'" % (self.selector, r, unit)

  def rotateY(self, r, unit='deg'):
    """
    Description:
    ------------
    The rotateY() CSS function defines a transformation that rotates an element around the ordinate (vertical axis) without deforming it.
    Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/rotateY

    Attributes:
    ----------
    :param r: Is an <angle> representing the angle of the rotation. A positive angle denotes a clockwise rotation, a negative angle a counter-clockwise one.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'rotateY(%s%s)'" % (self.selector, r, unit)

  def rotateZ(self, r, unit='deg'):
    """
    Description:
    ------------
    The rotateZ() CSS function defines a transformation that rotates an element around the z-axis without deforming it. Its result is a <transform-function> data type.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/rotateZ

    Attributes:
    ----------
    :param r: Is an <angle> representing the angle of the rotation. A positive angle denotes a clockwise rotation, a negative angle a counter-clockwise one.
    :param unit: String. The unit for the transformation (deg, turn, rad...)
    """
    return "%s.style.transform = 'rotateZ(%s%s)'" % (self.selector, r, unit)


class JsDoms(JsObject.JsObject):
  _id = None
  display_value = 'block'

  @classmethod
  def new(cls, tagName=None, varName=None, isPyData=True, setVar=True, report=None):
    """
    Description:
    ------------
    Create a new dom object to be added to the HTML page

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_date.asp

    Attributes:
    ----------
    :param tagName: The tag name to be created
    :param varName: Optional,
    :param isPyData: Optional,

    :return: The Python Javascript Date primitive
    """
    return cls(data="document.createElement('%s')" % tagName, varName=varName, setVar=setVar, isPyData=isPyData, report=report)

  def querySelector(self, tag):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param tag:

    :return:
    """
    return JsDoms("%s.querySelector('%s')" % (self.toStr(), tag))

  def querySelectorAll(self, tag, varName=None):
    """
    Description:
    ------------
    The querySelectorAll() method returns all elements in the document that matches a specified CSS selector(s), as a static NodeList object.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_queryselectorall.asp

    Attributes:
    ----------
    :param tag: String. The tag name
    :param varName: String. The variable name on the javascript side

    :return: A javascript Array object
    """
    if varName is None:
      return JsArray.JsArray("%s.querySelectorAll('%s')" % (self.toStr(), tag), isPyData=False)

    return JsArray.JsArray("%s.querySelectorAll('%s')" % (self.toStr(), tag), varName=varName, setVar=True, isPyData=False)

  def empty(self):
    """
    Description:
    ------------
    Shortcup function to emtpy an HTML compnent.
    This will only reuse the innerHTML property
    """
    return JsObject.JsObject('%s.innerHTML = ""' % (self.varId))

  @property
  def transform(self):
    """
    Description:
    ------------
    The transform property applies a 2D or 3D transformation to an element.
    This property allows you to rotate, scale, move, skew, etc., elements.

    Related Pages:

			https://www.w3schools.com/jsref/prop_style_transform.asp
    https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function
    """
    return JsDomsTransforms(self._report, self.varId)

  @property
  def jquery(self):
    """
    Description:
    ------------
    Link to the Jquery package

    THe id attribute must be defined
    """
    from epyk.core.js.packages import JsQuery

    if self._id is None:
      raise Exception("Id must be defined to attach Jquery features to this object")

    if getattr(self, '_jq', None) is None:
      self._jq = JsQuery.JQuery(self._report, selector="jQuery('#%s')" % self._id, setVar=False)
    return self._jq

  def addOnReady(self, jsFncs):
    """
    Description:
    ------------
    The ready event occurs when the DOM (document object model) has been loaded.

    Related Pages:

			https://www.w3schools.com/jquery/event_ready.asp

    Attributes:
    ----------
    :param jsFncs: The Javascript functions to be added to this section
    """
    self._report._props.setdefault('js', {}).setdefault('onCompReady', {})[self.varId] = ";".join(JsUtils.jsConvertFncs(jsFncs))

  def innerText(self, jsString=None, append=False, valType=None):
    """
    Description:
    ------------
    The innerText property sets or returns the text content of the specified node, and all its descendants.

    Usage::

      select.label.dom.innerText("test Change")

    Related Pages:

			https://www.w3schools.com/jsref/prop_node_innertext.asp

    Attributes:
    ----------
    :param jsString: Optional, The Javascript String to be added
    :param append: Boolean. Mention if the component should replace or append the data
    :param valType: Type: The type of data expected in the component

    :return: The JsObj to allow the chaining
    """
    if jsString is None:
      return "%s.innerText" % self.varId

    if append:
      if valType == int:
        self._js.append("%s.innerText = parseInt(%s.innerText) + %s" % (self.varId, self.varId, JsUtils.jsConvertData(jsString, None)))
      elif valType == float:
        self._js.append("%s.innerText = parseFloat(%s.innerText) + %s" % (self.varId, self.varId, JsUtils.jsConvertData(jsString, None)))
      else:
        self._js.append("%s.innerText += %s" % (self.varId, JsUtils.jsConvertData(jsString, None)))
    else:
      self._js.append("%s.innerText = %s" % (self.varId, JsUtils.jsConvertData(jsString, None)))
    return self

  def innerHTML(self, jsString=None, append=False, valType=None):
    """
    Description:
    ------------
    Sets or returns the content of an element

    Usage::

      select.label.dom.innerHTML("<p style='color:red'>Changed !</p>")

    Related Pages:

			https://www.w3schools.com/jsref/prop_html_innerhtml.asp

    Attributes:
    ----------
    :param jsString: Optional, The Javascript String to be added
    :param append: Boolean. Mention if the component should replace or append the data
    :param valType: Type: The type of data expected in the component

    :return: The JsObj
    """
    if jsString is None:
      return JsString.JsString("%s.innerHTML" % self.varId, isPyData=False)

    if append:
      if valType == int:
        self._js.append("%s.innerHTML = parseInt(%s.innerHTML) + %s" % (self.varId, self.varId, JsUtils.jsConvertData(jsString, None)))
      elif valType == int:
        self._js.append("%s.innerHTML = parseFloat(%s.innerHTML) + %s" % (self.varId, self.varId, JsUtils.jsConvertData(jsString, None)))
      else:
        self._js.append("%s.innerHTML += %s" % (self.varId, JsUtils.jsConvertData(jsString, None)))
    else:
      self._js.append("%s.innerHTML = %s" % (self.varId, JsUtils.jsConvertData(jsString, None)))
    return self

  def attr(self, type, jsObject=None):
    """
    Description:
    -----------
    The attr() method adds the specified attribute to an element, and gives it the specified value.
    It will use the underlying setAttribute() method

    Usage::

      select.label.dom.attr("title", "Tooltip")
    select.label.dom.attr({"title": "Tooltip"})

    Related Pages:

			https://www.w3schools.com/jsref/met_element_setattribute.asp

    Attributes:
    ----------
    :param type: A String with the type of parameter or a python dictionary
    :param jsObject: A JsObj with the value to be set

    :return: A JsObj
    """
    if jsObject is None and isinstance(type, dict):
      for k, v in type.items():
        if k == "id":
          self._id = v
        self._js.append("%s.setAttribute('%s', %s)" % (self.varId, k, JsUtils.jsConvertData(v, None)))
    else:
      if type == "id":
        self._id = jsObject
      self._js.append("%s.setAttribute('%s', %s)" % (self.varId, type, JsUtils.jsConvertData(jsObject, None)))
    return self

  def setAttribute(self, attributename, attributevalue):
    """
    Description:
    -----------
    The setAttribute() method adds the specified attribute to an element, and gives it the specified value.

    Usage::

      select.label.dom.setAttribute("title", "Tooltip")

    Related Pages:

			https://www.w3schools.com/jsref/met_element_setattribute.asp

    Attributes:
    ----------
    :param attributename: Required. The name of the attribute you want to add
    :param attributevalue: Required. The value of the attribute you want to add
    """
    self._js.append("%s.setAttribute('%s', %s)" % (self.varId, attributename, JsUtils.jsConvertData(attributevalue, None)))
    return self

  def addClass(self, clsName, attrs=None, eventAttrs=None, extend=True):
    """
    Description:
    -----------
    Adds the specified class(es) to each element in the set of matched elements.

    This function can either use an existing class or create one if the attrs or eventAttrs are defined

    Usage::

      table.dom.addClass("red", {"border": "1px solid green"}, extend=False)

    Related Pages:

			https://www.w3schools.com/jsref/met_element_setattribute.asp

    Attributes:
    ----------
    :param clsName: The Css classname
    :param attrs: A python dictionary with the css attributes
    :param eventAttrs: A nested python dictionary with the css attributes for each events
    :param extend: Boolean. To set if the class should replace the existing style definition

    :return:
    """
    if attrs is not None or eventAttrs is not None:
      clsName = self._report.style.cssName(clsName)
      self._report.style.cssCls(clsName, attrs, eventAttrs, False)
    if extend:
      self._js.append('%s.setAttribute("class", %s.getAttribute("class") + " %s")' % (self.varId, self.varId, clsName))
    else:
      self._js.append('%s.setAttribute("class", "%s")' % (self.varId, clsName))
    return self

  def css(self, type, jsObject=None):
    """
    Description:
    -----------
    Replicate in plain Js the Jquery CSS function

    Usage::

      select.label.dom.css({"color": "red"})

    Related Pages:

			https://www.w3schools.com/jsref/met_element_setattribute.asp

    Attributes:
    ----------
    :param type: A String with the type of parameter or a python dictionary
    :param jsObject: A JsObj with the value to be set

    :return: A JsObj
    """
    if jsObject is None and isinstance(type, dict):
      for k, v in type.items():
        if "-" in k:
          split_css = k.split("-")
          k = "%s%s" % (split_css[0], split_css[1].title())
        self._js.append("%s.style.%s = %s" % (self.varId, k, JsUtils.jsConvertData(v, None)))
    elif jsObject is None:
      if "-" in type:
        split_css = type.split("-")
        type = "%s%s" % (split_css[0], split_css[1].title())
      return JsObject.JsObject("%s.style.%s" % (self.varId, type))
    else:
      if "-" in type:
        split_css = type.split("-")
        type = "%s%s" % (split_css[0], split_css[1].title())
      self._js.append("%s.style.%s = %s" % (self.varId, type, JsUtils.jsConvertData(jsObject, None)))
    return self

  def position(self, x=None, y=None, dx=0, dy=0):
    """
    Description:
    -----------
    Set the position of the component in the page.
    By default the component will be fixed at the mouse level (this should be used in an event).

    Usage::

      rptObj.js.createElement("div", "popup").innerHTML("uygk,k,kj..kj.kjyf").attr('id', 'popup').css({
        'color': 'red', 'display': 'block'}).position()

    Attributes:
    ----------
    :param x: Integer. The positinn from the top of the page
    :param y: Integer. The position from the left
    :param dx: Integer.
    :param dy: Integer.

    :return: A JsObj
    """
    if x is None and y is None:
      self.css({"position": 'absolute', 'top': JsObject.JsObject.get("(event.clientY + window.scrollY + %s) + 'px'" % dy),
                'left': JsObject.JsObject.get("(event.clientX + window.scrollX + %s) + 'px'" % dx)})
    else:
      self.css({"position": 'absolute', 'top': "%spx" % x or 0, 'left': "%spx" % x or 0})
    return self

  def toggle_transition(self, attribute, value, value2, duration=1, timing_fnc='ease', delay=None):
    """
    Description:
    ------------
    Toogle a transition

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_transition.asp

    Attributes:
    ----------
    :param attribute: String or List. Specifies the name of the CSS property the transition effect is for
    :param value: String or List. Specifies the value of the CSS property the transition effect is for
    :param value2: String or List. Specifies the value of the CSS property the transition effect is for
    :param duration: Number or List. Specifies how many seconds or milliseconds the transition effect takes to complete
    :param delay: Number. Defines when the transition effect will start
    :param timing_fnc: String. The transition-timing-function property specifies the speed curve of the transition effect.
    """
    if "-" in attribute:
      split_css = attribute.split("-")
      css_attr = "%s%s" % (split_css[0], split_css[1].title())
    else:
      css_attr = attribute
    tmp = list(self._js)
    self._js = []
    self.transition(attribute, value, duration, delay, timing_fnc)
    if_ = "; ".join(self._js)

    self._js = []
    self.transition(attribute, value2, duration, delay, timing_fnc)
    else_ = "; ".join(self._js)
    self._js = tmp
    self._js.append("if(%s.style.%s != '%s') {%s} else {%s}" % (self.varId, css_attr, value, if_, else_))
    return self

  def transition(self, attribute, value, duration=1, delay=None, timing_fnc='ease', reverse=False):
    """
    Description:
    ------------
    The transition property is a shorthand property for:
      - transition-property
      - transition-duration
      - transition-timing-function
      - transition-delay

    Usage::

      i.label.dom.transition('margin-left', '100px', 2, reverse=True),
    i.label.dom.transition('color', 'red', 5, reverse=True),

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_transition.asp

    Attributes:
    ----------
    :param attribute: String or List. Specifies the name of the CSS property the transition effect is for
    :param value: String or List. Specifies the value of the CSS property the transition effect is for
    :param duration: Number or List. Specifies how many seconds or milliseconds the transition effect takes to complete
    :param delay: Number. Defines when the transition effect will start
    :param timing_fnc: String. The transition-timing-function property specifies the speed curve of the transition effect.
    :param reverse: Boolean. Rewind the transition animation
    """
    self.css("transition-property", ",".join(attribute) if isinstance(attribute, list) else attribute)
    if isinstance(duration, list):
      self.css("transition-duration", "%ss" % "s, ".join(map(lambda x: str(x), duration)))
    else:
      self.css("transition-duration", "%ss" % duration)
    self.css("transition-timing-function", timing_fnc)
    if delay is not None:
      if isinstance(delay, int):
        self.css("transition-delay", "%ss" % delay)
      else:
        self.css("transition-delay", delay)
    if reverse:
      if isinstance(attribute, list):
        for i, attr in enumerate(attribute):
          self._js.append(self.css(attr).setVar("css_transition_%s" % i).r)
          self._js.append("setTimeout(function(){%s = css_transition_%s}, %s)" % (self.css(attr), i, duration[i] * 1000))
      else:
        self._js.append(self.css(attribute).setVar("css_transition").r)
        self._js.append("setTimeout(function(){%s = css_transition}, %s)" % (self.css(attribute), duration * 1000))

    if isinstance(attribute, list):
      for i, attr in enumerate(attribute):
        self.css(attr, value[i])
    else:
      self.css(attribute, value)
    self.css("transition-property", "initial")
    return self

  def hide(self):
    """
    Description:
    ------------

    Usage::

      input.js.hide()

    Related Pages:

			https://gomakethings.com/how-to-show-and-hide-elements-with-vanilla-javascript/
    """
    return self.css("display", "none")

  def show(self):
    """
    Description:
    ------------

    Usage::

      input.js.hide()

    Related Pages:

			https://gomakethings.com/how-to-show-and-hide-elements-with-vanilla-javascript/
    """
    return self.css("display", self.display_value)

  def toggle(self, attr="display", jsVal1=None, jsVal2="none"):
    """
    Description:
    ------------
    Hexadecimal colors should be converted to rgb code as only the computed style will be compared.
    To do so you can use: Colors.getHexToRgb(self._report.theme.success[1]) from epyk.core.css import Colors

    Attributes:
    ----------
    :param attr:
    :param jsVal1:
    :param jsVal2:
    """
    if attr == 'display' and jsVal1 is None:
      jsVal1 = self.display_value
    if "-" in attr:
      split_css = attr.split("-")
      attr = "%s%s" % (split_css[0], split_css[1].title())
    self._js.append("if(window.getComputedStyle(%(varId)s).%(attr)s == '%(jsVal1)s'){ %(varId)s.style.%(attr)s = '%(jsVal2)s'} else { %(varId)s.style.%(attr)s = '%(jsVal1)s'}" % {"varId": self.varId, "attr": attr, "jsVal1": jsVal1, "jsVal2": jsVal2})
    return self

  def toggleAttrs(self, pivot_key, pivot_val, attrs_off, attrs_on):
    """
    Description:
    ------------
    Toggle some CSS attributes

    Attributes:
    ----------
    :param pivot_key:
    :param pivot_val:
    :param attrs_on: A python dictionary with CSS attributes
    :param attrs_off: A python dictionary with CSS attributes
    """
    if pivot_key in ["color"] and not pivot_val.startswith("rgb"):
      colors_def = Colors.defined[pivot_val.upper()]
      pivot_val = "rgb%s" % colors_def['rgb']
    css_attrs_on = self.css(attrs_on).toStr()
    css_attrs_off = self.css(attrs_off).toStr()
    self._js.append("if(window.getComputedStyle(%(varId)s)['%(pivot_key)s'] == '%(pivot_val)s') {%(css_attrs_on)s} else {%(css_attrs_off)s}" % {"pivot_val": pivot_val, "varId": self.varId, "pivot_key": pivot_key, 'css_attrs_on': css_attrs_on, 'css_attrs_off': css_attrs_off})
    return self

  def toggleText(self, jsString1, jsString2):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsString:
    """
    str1 = JsUtils.jsConvertData(jsString1, None)
    str2 = JsUtils.jsConvertData(jsString2, None)
    self._js.append("if(%(varId)s.innerText == %(str2)s) {%(varId)s.innerText = %(str1)s} else {%(varId)s.innerText = %(str2)s}" % {"varId": self.varId, "str1": str1, "str2": str2})
    return self

  @property
  def clientHeight(self):
    """
    Description:
    ------------
    The Element.clientHeight read-only property is zero for elements with no CSS or inline layout boxes; otherwise, it's the inner height of an element in pixels.
    It includes padding but excludes borders, margins, and horizontal scrollbars (if present).

    Usage::

      rptObj.js.alert(rptObj.js.body.clientHeight)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Element/clientHeight
    """
    return JsNumber.JsNumber("%s.clientHeight" % self.varId)

  def toggleClass(self, clsName, propagate=False):
    """
    Description:
    ------------
    Toggle a class name

    Attributes:
    ----------
    :param clsName: The classname to be toggle

    :return:
    """
    clsName = JsUtils.jsConvertData(clsName, None)
    if propagate:
      self._js.append('%(varId)s.parentNode.childNodes.forEach(function(e){e.classList.remove(%(data)s)})' % {"varId": self.varId, 'data': clsName})
    self._js.append('var classes = %(data)s.split(" "); classes.forEach(function(cls){ %(varId)s.classList.toggle(cls); })' % {"varId": self.varId, 'data': clsName})
    return self

  def switchClass(self, clsName1, clsName2):
    """
    Description:
    ------------
    Switch from one CSS class to another.

    Usage::

      icon.dom.switchClass("fa-folder", "fa-folder-open")

    Attributes:
    ----------
    :param clsName1: String. A class name or a string with a list of classname space separated
    :param clsName2: String. A class name or a string with a list of classname space separated

    :return: Self to allow the chaining
    """
    self.toggleClass(clsName1)
    self.toggleClass(clsName2)
    return self

  @property
  def firstChild(self):
    """
    Description:
    ------------
    The firstChild property returns the first child node of the specified node, as a Node object.

    Usage::

      select.dom.firstChild
    select.dom.firstChild.css({"color": "yellow"})

    Related Pages:

			https://www.w3schools.com/jsref/prop_node_firstchild.asp

    :return: A new JsDom python object
    """
    return JsDoms("%s.firstChild" % self.varId)

  @property
  def nextSibling(self):
    """
    Description:
    ------------
    The nextSibling property returns the node immediately following the specified node, in the same tree level.

    Related Pages:

			https://www.w3schools.com/jsref/prop_node_nextsibling.asp
    """
    return JsDoms("%s.nextSibling" % self.varId)

  def contains(self, node):
    """
    Description:
    ------------
    The contains() method returns a Boolean value indicating whether a node is a descendant of a specified node.

    Related Pages:

			https://www.w3schools.com/jsref/met_node_contains.asp

    Attributes:
    ----------
    :param node: Required. Specifies the node that may be contained by (a descendant of) a specified node

    :return: A Boolean
    """
    return JsBoolean.JsBoolean('%s.contains(%s)' % (self.varId, node))

  def getAttribute(self, attributename):
    """
    Description:
    ------------
    The getAttribute() method returns the value of the attribute with the specified name, of an element.

    Usage::

      select.dom.getAttribute("class")

    Related Pages:

			https://www.w3schools.com/jsref/met_element_getattribute.asp

    Attributes:
    ----------
    :param attributename: Required. The name of the attribute you want to get the value from

    :return: A String, representing the specified attribute's value.
    """
    return JsObject.JsObject("%s.getAttribute(%s)" % (self.varId, JsUtils.jsConvertData(attributename, None)), isPyData=False)

  def getAttributeNode(self, attributename):
    """
    Description:
    ------------
    The getAttributeNode() method returns the attribute node with the specified name of an element, as an Attr object.

    Related Pages:

			https://www.w3schools.com/jsref/met_element_getattributenode.asp

    Attributes:
    ----------
    :param attributename: Required. The name of the attribute you want to return

    :return: An Attr object, representing the specified attribute node.
    """
    return JsString.JsString(varId="%s.getAttributeNode('%s')" % (self.varId, attributename))

  def getComputedStyle(self, attributename=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param attributename:
    """
    if attributename is None:
      return JsString.JsString("getComputedStyle(%s)" % self.varId, isPyData=False)

    if "-" in attributename:
      split_css = attributename.split("-")
      attributename = "%s%s" % (split_css[0], split_css[1].title())
    return JsString.JsString("getComputedStyle(%s).%s" % (self.varId, attributename), isPyData=False)

  def getBoundingClientRect(self):
    """
    Description:
    -----------
    The getBoundingClientRect() method returns the size of an element and its position relative to the viewport.

    Related Pages:

			https://www.w3schools.com/jsref/met_element_getboundingclientrect.asp
    https://developer.mozilla.org/en-US/docs/Web/API/DOMRect
    """
    return JsNodeDomRect.JsDOMRect("%s.getBoundingClientRect()" % self.varId)

  @property
  def hasChildNodes(self):
    """
    Description:
    ------------
    Returns true if an element has any child nodes, otherwise false

    Usage::

      select.dom.hasChildNodes

    Related Pages:

			https://www.w3schools.com/jsref/met_node_haschildnodes.asp

    :return: A Boolean, returns true if the node has child nodes, false otherwise
    """
    return JsBoolean.JsBoolean("%s.hasChildNodes()" % self.varId, isPyData=False)

  def hasClass(self, className):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param className:

    :return:
    """
    className = JsUtils.jsConvertData(className, None)
    return JsBoolean.JsBoolean("%s.classList.contains(%s)" % (self.varId, className), isPyData=False)

  def text(self, jsString):
    """
    Description:
    ------------
    Javascript Function

    Shortcut in charge oa creating a text node object and adding the text.

    Related Pages:
Attributes:
    ----------
    :param jsString: The Javascript String of the text node component

    :return: The main Python Dom Object
    """
    return self.appendChild(JsFncs.JsFunction("document.createTextNode(%s)" % JsUtils.jsConvertData(jsString, None)))

  @property
  def childNodes(self):
    """
    Description:
    ------------
    The childNodes property returns a collection of a node's child nodes, as a NodeList object.

    The nodes in the collection are sorted as they appear in the source code and can be accessed by index numbers. The index starts at 0.

    Related Pages:

			https://www.w3schools.com/jsref/prop_node_childnodes.asp

    :return: A NodeList object, representing a collection of nodes. The nodes in the returned collection are sorted as they appear in the source code
    """
    self._js.append("%s.childNodes" % self.varId)
    return self

  @property
  def tagName(self):
    """
    Description:
    ------------
    The tagName property returns the tag name of the element

    Usage::

      select.dom.tagName

    Related Pages:

			https://www.w3schools.com/jsref/prop_element_tagname.asp

    :return: A String, representing the tag name of the element in uppercase
    """
    return JsString.JsString("%s.tagName" % self.varId, isPyData=False)

  def contentEditable(self, bool):
    """
    Description:
    ------------
    Set content editable

    Usage::

      rptObj.ui.text("This is a text").dom.contentEditable(True)

    Attributes:
    ----------
    :param bool: Boolean. Set the content editable flag to the Dom object

    :return: Return a JsBoolean object
    """
    return JsBoolean.JsBoolean.get("%s.contentEditable = %s" % (self.varId, JsUtils.jsConvertData(bool, None)))

  def className(self, className=None):
    """
    Description:
    ------------
    The className property sets or returns the class name of an element (the value of an element's class attribute).

    Usage::

      select.dom.className()

    Related Pages:

			https://www.w3schools.com/jsref/prop_html_classname.asp

    Attributes:
    ----------
    :param className: Specifies the class name of an element. To apply multiple classes, separate them with spaces, like "test demo"

    :return: A String, representing the class, or a space-separated list of classes, of an element
    """
    if className is None:
      return JsString.JsString("%s.className" % self.varId, isPyData=False)

    # TODO fix this properly
    return JsString.JsString("%s; %s.className = %s" % (self.toStr(), self.varId, JsUtils.jsConvertData(className, None)), isPyData=False)

  def cloneNode(self, deep=True):
    """
    Description:
    ------------
    The cloneNode() method creates a copy of a node, and returns the clone.

    The cloneNode() method clones all attributes and their values.

    Usage::

      select.dom.cloneNode()

    Related Pages:

			https://www.w3schools.com/jsref/met_node_clonenode.asp

    Attributes:
    ----------
    :param deep: Optional. Specifies whether all descendants of the node should be cloned.

    :return: A Node object, representing the cloned node
    """
    return JsDoms("%s.cloneNode(%s)" % (self.varId, JsUtils.jsConvertData(deep, None)))

  def remove(self):
    """
    Description:
    ------------
    Remove the current dom object from the page

    Usage::

      select.dom.remove()

    Related Pages:

			https://developer.mozilla.org/fr/docs/Web/API/ChildNode/remove
    """
    return JsFncs.JsFunction("%s.remove()" % self.varId)

  def removeChild(self, jsDom):
    """
    Description:
    ------------
    Removes a child node from an element

    Related Pages:

			https://www.w3schools.com/jsref/met_node_removechild.asp

    Attributes:
    ----------
    :param jsDom: Required. The node object you want to remove

    :return: A Node object, representing the removed node, or null if the node does not exist
    """
    return JsDoms("%s.removeChild(%s)" % (self.varId, jsDom))

  def appendChild(self, jsDom):
    """
    Description:
    ------------
    The appendChild() method appends a node as the last child of a node.

    Usage::

      select.dom.appendChild(select.label.dom.cloneNode())

    Related Pages:

			https://www.w3schools.com/jsref/met_node_appendchild.asp

    Attributes:
    ----------
    :param jsDom: Required. The node object you want to append

    :return: 	A Node Object, representing the appended node
    """
    self._js.append("%s.appendChild(%s)" % (self.varId, JsUtils.jsConvertData(jsDom, None)))
    return self

  def insertBefore(self, newnode, existingnode=None):
    """
    Description:
    ------------
    The insertBefore() method inserts a node as a child, right before an existing child, which you specify.

    Usage::

      select.dom.insertBefore(select.label.dom.cloneNode())

    Related Pages:

			https://www.w3schools.com/jsref/met_node_insertbefore.asp

    Attributes:
    ----------
    :param newnode: Required. The node object you want to insert
    :param existingnode: Optional. The child node you want to insert the new node before. If set to null, the insertBefore method will insert the newnode at the end
    """
    if existingnode is None:
      self._js.append("%s.insertBefore(%s, %s)" % (self.varId, newnode, self.firstChild))
    else:
      self._js.append("%s.insertBefore(%s, %s)" % (self.varId, newnode, existingnode))
    return self

  def click(self, jsFncs=None):
    """
    Description:
    -----------
    Trigger a click event.
    This function will not set the event

    Attributes:
    ----------
    :param jsFncs:
    """
    if jsFncs is None:
      return JsObject.JsObject("%s.click()" % self.varId)

    self._js.append("%s.click(%s)" % (self.varId, ";".join(JsUtils.jsConvertFncs(jsFncs))))
    return self

  def onclick(self, jsFncs, autoStyle=True):
    """
    Description:
    -----------
    Execute a JavaScript when a button is clicked

    Related Pages:

			https://www.w3schools.com/jsref/event_onclick.asp

    Attributes:
    ----------
    :param jsFncs: The Javascript function
    :param autoStyle: Some predefined style attributes added to this event (self.css({"cursor": "pointer"}))

    :return: The PyDom object
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if autoStyle:
      self.css({"cursor": "pointer"})
    self._js.append("%s.onclick = function(){%s}" % (self.varId, ";".join(JsUtils.jsConvertFncs(jsFncs))))
    return self

  def onVisible(self, jsFncs):
    """
    Description:
    ------------

    :param jsFncs:

    :return:
    """
    self._js.append("var rect = elm.getBoundingClientRect()" % self.varId)
    return self

  def getContext(self, contextType, contextAttributes=None):
    """
    Function dedicated to DOM Canvas types.

    The HTMLCanvasElement.getContext() method returns a drawing context on the canvas, or null if the context identifier is not supported.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/HTMLCanvasElement/getContext

    Attributes:
    ----------
    :param contextType: Is a DOMString containing the context identifier defining the drawing context associated to the canvas
    :param contextAttributes: Dictionary with specific context attributes (depending on the type

    TODO: Add a check on the tag

    :return:
    """
    types = ["2d", "webgl", "experimental-webgl", "webgl2", "bitmaprenderer"]
    if contextType not in types:
      raise Exception("Context type %s not recognised" % contextType)

    if contextAttributes is None:
      return JsFncs.JsFunction("%s.getContext('%s')" % (self.varId, contextType))

    contextAttributes = JsUtils.jsConvertData(contextAttributes, None)
    return JsFncs.JsFunction("%s.getContext('%s', %s)" % (self.varId, contextType, contextAttributes))


class JsDomsList(JsArray.JsArray):

  def all(self, jsFncs):
    """
    Description:
    ------------
    Apply a set of functions on all the elements with this name.

    Attributes:
    ----------
    :param jsFncs: Array. List of Javascript fragments
    """
    self._js.append("%s.forEach(function(elt, index){%s})" % (self.varId, JsUtils.jsConvertFncs(jsFncs, toStr=True)))
    return self

  @property
  def first(self):
    """
    Description:
    ------------
    Get the first dom item in corresponding to the name criteria
    """
    return JsDoms.get("%s[0]" % self.toStr())

  def css(self, type, jsObject=None):
    """
    Description:
    ------------
    Replicate in plain Js the Jquery CSS function

    Usage::

      select.label.dom.css({"color": "red"})

    Related Pages:

			https://www.w3schools.com/jsref/met_element_setattribute.asp

    Attributes:
    ----------
    :param type: A String with the type of parameter or a python dictionary
    :param jsObject: A JsObj with the value to be set

    :return: A JsObj
    """
    if jsObject is None and isinstance(type, dict):
      for k, v in type.items():
        if "-" in k:
          split_css = k.split("-")
          k = "%s%s" % (split_css[0], split_css[1].title())
        self._js.append("for(let e of %s){ e.style.%s = %s }" % (self.varId, k, JsUtils.jsConvertData(v, None)))
    elif jsObject is None:
      if "-" in type:
        split_css = type.split("-")
        type = "%s%s" % (split_css[0], split_css[1].title())
      return JsObject.JsObject("for(let e of %s){ e.style.%s }" % (self.varId, type))
    else:
      if "-" in type:
        split_css = type.split("-")
        type = "%s%s" % (split_css[0], split_css[1].title())
      self._js.append("for(let e of %s){ e.style.%s = %s }" % (self.varId, type, JsUtils.jsConvertData(jsObject, None)))
    return self

  def attr(self, type, jsObject=None):
    """
    Description:
    ------------
    The attr() method adds the specified attribute to an element, and gives it the specified value.
    It will use the underlying setAttribute() method

    Usage::

      select.label.dom.attr("title", "Tooltip")
    select.label.dom.attr({"title": "Tooltip"})

    Related Pages:

			https://www.w3schools.com/jsref/met_element_setattribute.asp

    Attributes:
    ----------
    :param type: A String with the type of parameter or a python dictionary
    :param jsObject: A JsObj with the value to be set

    :return: A JsObj
    """
    if jsObject is None and isinstance(type, dict):
      for k, v in type.items():
        if k == "id":
          self._id = v
        self._js.append("for(let e of %s){ e.setAttribute('%s', %s) }" % (self.varId, k, JsUtils.jsConvertData(v, None)))
    else:
      if type == "id":
        self._id = jsObject
      self._js.append("for(let e of %s){ e.setAttribute('%s', %s) }" % (self.varId, type, JsUtils.jsConvertData(jsObject, None)))
    return self

  def __getitem__(self, index):
    """
    Description:
    ------------
    Get the nth item corresponding to the name in the HTML page

    Attributes:
    ----------
    :param index: Integer. The index number of the item
    """
    return JsDoms.get("%s[%s]" % (self.toStr(), index))




from epyk.core.html import Html


class SVG(Html.Html):
  def __init__(self, report, width, height):
    super(SVG, self).__init__(report, "", css_attrs={"width": width, "height": height})
    self.origine = None
    if width is not None:
      self.set_attrs({"viewBox": "0 0 %s %s" % (width[0], height[0]), "version": '1.1', 'preserveAspectRatio': 'xMinYMin meet'})
    self.css({"display": 'inline-block'})
    self.html_objs = []

  def __add__(self, obj):
    obj.inReport = False
    self.html_objs.append(obj)

  def __getitem__(self, i):
    return self.html_objs[i]

  def defs(self):
    """
    Description:
    ------------
    The <defs> element is used to store graphical objects that will be used at a later time.
    Objects created inside a <defs> element are not rendered directly.
    To display them you have to reference them (with a <use> element for exampl

    Usage::

      defs = poly.defs()

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/defs
    """
    self.html_objs.append(Defs(self._report))
    return self.html_objs[-1]

  def text(self, text, x, y):
    """
    Description:
    ------------
    The <text> element can be arranged in any number of sub-groups with the <tspan> element.
    Each <tspan> element can contain different formatting and position.

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_text.asp

    Attributes:
    ----------
    :param text: String. The text to be added to the container
    :param x: Float. The x coordinate of the starting point of the text baseline.
    :param y: Float. The y coordinate of the starting point of the text baseline.

    :rtype: Text
    """
    self.html_objs.append(Text(self._report, text, x, y))
    return self.html_objs[-1]

  def rect(self, x, y, width, height, fill, rx=0, ry=0):
    """
    Description:
    ------------
    The <rect> element is used to create a rectangle and variations of a rectangle shape

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_rect.asp

    Attributes:
    ----------
    :param x:
    :param y:
    :param width:
    :param height:
    :param fill:
    :param rx:
    :param ry:
    """
    self.html_objs.append(Rectangle(self._report, x, y, width, height, fill, rx, ry))
    return self.html_objs[-1]

  def line(self, x1, y1, x2, y2):
    """
    Description:
    ------------
    The <line> element is used to create a line

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_line.asp

    Attributes:
    ----------
    :param x1: Float. The x1 attribute defines the start of the line on the x-axis
    :param y1: Float. The y1 attribute defines the start of the line on the y-axis
    :param x2: Float. The x2 attribute defines the end of the line on the x-axis
    :param y2: Float. The y2 attribute defines the end of the line on the y-axis

    :rtype: Line
    """
    self.html_objs.append(Line(self._report, x1, y1, x2, y2))
    return self.html_objs[-1]

  def circle(self, x, y, r, fill="None"):
    """
    Description:
    ------------
    The <circle> element is used to create a circle

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_circle.asp

    Attributes:
    ----------
    :param x: Float. The x coordinate of the circle
    :param y: Float. The y coordinate of the circle
    :param r: Float. The r attribute defines the radius of the circle

    :rtype: Circle
    """
    self.html_objs.append(Circle(self._report, x, y, r, fill))
    return self.html_objs[-1]

  def ellipse(self, cx, cy, rx, ry):
    """
    Description:
    ------------
    The <ellipse> element is used to create an ellipse

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_ellipse.asp

    Attributes:
    ----------
    :param cx: Float. The cx attribute defines the x coordinate of the center of the ellipse
    :param cy: Float. The cy attribute defines the y coordinate of the center of the ellipse
    :param rx: Float. The rx attribute defines the horizontal radius
    :param ry: Float. The ry attribute defines the vertical radius

    :rtype: Ellipse
    """
    self.html_objs.append(Ellipse(self._report, cx, cy, rx, ry))
    return self.html_objs[-1]

  def polygon(self, points, fill="None"):
    """
    Description:
    ------------
    The <polygon> element is used to create a graphic that contains at least three sides.

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_polygon.asp

    Attributes:
    ----------
    :param points: String. The points attribute defines the x and y coordinates for each corner of the polygon

    :rtype: Polygone
    """
    self.html_objs.append(Polygone(self._report, points, fill))
    return self.html_objs[-1]

  def polyline(self, points, fill="None"):
    """
    Description:
    ------------
    The <polyline> element is used to create any shape that consists of only straight lines (that is connected at several points)

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_polyline.asp

    Attributes:
    ----------
    :param points:

    :rtype: Polyline
    """
    polygone = Polyline(self._report, points, height=None, width=None, fill=fill, options={})
    self.html_objs.append(polygone)
    return polygone

  def triangle(self, points, fill="None", options=None):
    """
    Description:
    ------------
    A polyline element with three points

    Usage::

    Related Pages:

			https://www.w3schools.com/graphics/svg_polyline.asp

    Attributes:
    ----------
    :param points:
    :param options:

    :rtpye: Polyline
    """
    self.html_objs.append(Polyline(self._report, points, None, None, fill, options or {}))
    return self.html_objs[-1]

  def g(self, fill="None", stroke=None, stroke_width=None):
    """
    Description:
    ------------
    The <g> SVG element is a container used to group other SVG elements.

    Usage::

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/g

    Attributes:
    ----------
    :param fill: String. The color for the component background
    :param stroke: String. The color for the border
    :param stroke_width: Float. The width of the component's border

    :rtype: G
    """
    self.html_objs.append(G(self._report, fill, stroke, stroke_width))
    return self.html_objs[-1]

  def path(self, x=0, y=0, fill='none', from_origin=False, bespoke_path=None):
    """
    Description:
    ------------
    The <path> element is used to define a path.

    Related Pages:

			https://www.w3.org/TR/SVG/paths.html
      https://www.w3.org/TR/svg-paths/

    Attributes:
    ----------
    :param x:
    :param y:

    :rtype: Path
    """
    if from_origin:
      x += self.origine[0]
      y += self.origine[1]
    self.html_objs.append(Path(self._report, x, y, fill, self.origine, bespoke_path))
    return self.html_objs[-1]

  def foreignObject(self, x, y, width, height):
    """
    Description:
    ------------
    The <foreignObject> SVG element includes elements from a different XML namespace. In the context of a browser, it is most likely (X)HTML.

    Usage::

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/foreignObject

    Attributes:
    ----------
    :param x: Float. The x coordinate of the foreignObject.
    :param y: Float. The y coordinate of the foreignObject.
    :param width: Float or Percentage. The width of the foreignObject.
    :param height: Float or Percentage. The height of the foreignObject.

    :rtype: ForeignObject
    """
    self.html_objs.append(ForeignObject(self._report, x, y, width, height))
    return self.html_objs[-1]

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<svg %s>%s</svg>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class LinearGradient(Html.Html):
  def __init__(self, report, htmlCode, x1, y1, x2, y2, gradientTransform):
    super(LinearGradient, self).__init__(report, "", htmlCode=htmlCode)
    self.set_attrs({'gradientTransform': gradientTransform, "x1": x1, "y1": y1, "x2": x2, "y2": y2})
    self.items = []

  @property
  def url(self):
    return "url(#%s)" % self.htmlCode

  def stop(self, offset, styles):
    self.items.append('<stop offset="%s" style="%s" />' % (offset, ";".join(["%s:%s" % (k, v) for k, v in styles.items()])))
    return self

  def __str__(self):
    return "<linearGradient %s>%s</linearGradient>" % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(self.items))


class RadialGradient(Html.Html):
  def __init__(self, report, htmlCode):
    super(RadialGradient, self).__init__(report, "", htmlCode=htmlCode)
    self.items = []

  @property
  def url(self):
    return "url(#%s)" % self.htmlCode

  def stop(self, offset, styles):
    self.items.append('<stop offset="%s" style="%s" />' % (offset, ";".join(["%s:%s" % (k, v) for k, v in styles.items()])))
    return self

  def __str__(self):
    return "<radialGradient %s>%s</radialGradient>" % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(self.items))


class Marker(SVG):
  def __init__(self, report, htmlCode, viewBox, refX, refY):
    super(Marker, self).__init__(report, None, None)
    self.htmlCode = htmlCode
    self.set_attrs({'id': htmlCode, "viewBox": viewBox, "refX": refX, "refY": refY})
    self.html_objs = []

  @property
  def url(self):
    """

    https://developer.mozilla.org/en-US/docs/Web/SVG/Element/marker

    :return:
    """
    return "url(#%s)" % self.htmlCode

  def orient(self, orientation):
    """

    :param orientation:

    :return:
    """
    self.set_attrs(name="orient", value=orientation)
    return self

  def markerWidth(self, value):
    """

    :param value:
    :return:
    """
    self.set_attrs(name="markerWidth", value=value)
    return self

  def markerHeight(self, value):
    """

    :param value:

    :return:
    """
    self.set_attrs(name="markerHeight", value=value)
    return self

  def arrow(self, size=None):
    """

    https://developer.mozilla.org/en-US/docs/Web/SVG/Element/marker

    :return:
    """
    self.html_objs.append("<path d='M 0 0 L 10 5 L 0 10 z' />")
    if size is not None:
      self.markerWidth(size).markerHeight(size)
    return self

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<marker %s>%s</marker>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Defs(Html.Html):
  def __init__(self, report):
    super(Defs, self).__init__(report, "")
    self.html_objs = []

  def linearGradient(self, htmlCode, x1="0%", y1="0%", x2="100%", y2="0%", gradientTransform=None):
    """
    Description:
    ------------
    The <linearGradient> element lets authors define linear gradients that can be applied to fill or stroke of graphical elements.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/linearGradient

    Attributes:
    ----------
    :param htmlCode: String. The HTML id of the component
    :param x1: Float.
    :param y1: Float.
    :param x2: Float.
    :param y2: Float.
    :param gradientTransform:

    :rtype: LinearGradient
    """
    self.html_objs.append(LinearGradient(self._report, htmlCode, x1, y1, x2, y2, gradientTransform))
    return self.html_objs[-1]

  def radialGradient(self, htmlCode):
    """
    Description:
    ------------
    The <radialGradient> element lets authors define radial gradients that can be applied to fill or stroke of graphical elements.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/radialGradient

    Attributes:
    ----------
    :param htmlCode: String. The HTML id of the component

    :rtype: radialGradient
    """
    self.html_objs.append(RadialGradient(self._report, htmlCode))
    return self.html_objs[-1]

  def marker(self, htmlCode, viewBox, refX, refY):
    """
    The <marker> element defines the graphic that is to be used for drawing arrowheads or polymarkers on a given <path>, <line>, <polyline> or <polygon> element.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/marker

    Attributes:
    ----------
    :param htmlCode:
    :param viewBox:
    :param refX:
    :param refY:

    :rtype: Marker
    """
    self.html_objs.append(Marker(self._report, htmlCode, viewBox, refX, refY))
    return self.html_objs[-1]

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<defs %s>%s</defs>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class ForeignObject(SVG):
  def __init__(self, report, x, y, width, height):
    super(ForeignObject, self).__init__(report, None, None)
    self.set_attrs({"x": x, "y": y, "width": width, "height": height})
    self.html_objs = []

  def add(self, htmlObj):
    """

    :param htmlObj:
    """
    if not isinstance(htmlObj, list):
      htmlObj = [htmlObj]
    for h in htmlObj:
      h.inReport = False
      self.html_objs.append(h)

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<foreignObject %s>%s</foreignObject>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class G(SVG):
  def __init__(self, report, fill, stroke, stroke_width):
    super(G, self).__init__(report, None, None)
    self.set_attrs({"fill": fill, "stroke": stroke, "stroke-width": stroke_width})
    self.html_objs = []

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<g %s>%s</g>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class SVGItem(Html.Html):

  def fill(self, color):
    """

    :param color:
    :return:
    """
    self.set_attrs(name="fill", value=color)
    return self

  def transform(self, attributeName, type, from_pos, to_pos, duration=4, repeatCount="indefinite"):
    """

    :param attributeName:
    :param type:
    :param from_pos:
    :param to_pos:
    :param duration:
    :param repeatCount:

    :return:
    """
    self.html_objs.append(AnimateTransform(self._report, attributeName, type, from_pos, to_pos, duration, repeatCount))
    return self


class Polygone(SVGItem):
  def __init__(self, report, points, fill):
    super(Polygone, self).__init__(report, points)
    self.set_attrs(({"points": " ".join(["%s,%s" % (x, y) for x, y in self.val]), "fill": fill}))
    self.css({'stroke': report.theme.greys[-1], 'stroke-width': 1})
    self.html_objs = []

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<polygon %s>%s</polygon>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Ellipse(SVGItem):
  def __init__(self, report, cx, cy, rx, ry):
    super(Ellipse, self).__init__(report, "")
    self.set_attrs({"cx": cx, "cy": cy, "rx": rx, "ry": ry})
    self.css({'stroke': report.theme.success[1], 'stroke-width': 1, 'fill': 'none'})
    self.html_objs = []

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<ellipse %s>%s</ellipse>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Line(SVGItem):
  def __init__(self, report, x1, y1, x2, y2):
    super(Line, self).__init__(report, "")
    self.set_attrs({"x1": x1, "y1": y1, "x2": x2, "y2": y2})
    self.css({"stroke": report.theme.greys[-1], "stroke-width": 1})
    self.html_objs = []

  def markers(self, marker_code):
    """

    :param marker_code:
    """
    self.set_attrs(name="marker-start", value=marker_code)
    self.set_attrs(name="marker-mid", value=marker_code)
    self.set_attrs(name="marker-end", value=marker_code)
    return self

  def marker_start(self, marker_code):
    self.set_attrs(name="marker-start", value=marker_code)

  def marker_mid(self, marker_code):
    self.set_attrs(name="marker-mid", value=marker_code)

  def marker_end(self, marker_code):
    self.set_attrs(name="marker-end", value=marker_code)

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<line %s>%s</line>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Polyline(SVGItem):
  def __init__(self, report, points, height, width, fill, options):
    super(Polyline, self).__init__(report, points, css_attrs={"width": width, "height": height})
    self.set_attrs({"fill": fill, "points": " ".join(["%s,%s" % (x, y) for x, y in self.val])})
    self.html_objs = []
    if options is not None:
      self._jsStyles.update(options)
    self.css({"display": 'inline-block', "fill": options.get('fill', ''),
              'stroke': options.get('stroke', report.theme.success[1]), 'stroke-width': options.get('stroke-width', 1)})

  def markers(self, marker_code):
    """

    :param marker_code:
    """
    self.set_attrs(name="marker-start", value=marker_code)
    self.set_attrs(name="marker-mid", value=marker_code)
    self.set_attrs(name="marker-end", value=marker_code)
    return self

  def marker_start(self, marker_code):
    self.set_attrs(name="marker-start", value=marker_code)

  def marker_mid(self, marker_code):
    self.set_attrs(name="marker-mid", value=marker_code)

  def marker_end(self, marker_code):
    self.set_attrs(name="marker-end", value=marker_code)

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<polyline %s>%s</polyline>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Rectangle(SVGItem):
  def __init__(self, report, x, y, width, height, fill, rx, ry):
    super(Rectangle, self).__init__(report, "", css_attrs={"width": width, "height": height})
    self.set_attrs({"x": x, "y": y, "fill": fill, "rx": rx, "ry": ry})
    self.css({"display": 'inline-block'})
    self.html_objs = []

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<rect %s>%s</rect>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Circle(SVGItem):
  def __init__(self, report, x, y, r, fill):
    super(Circle, self).__init__(report, "")
    self.set_attrs({"cx": x, "cy": y, "r": r, "fill": fill})
    self.html_objs = []

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<circle %s>%s</circle>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_c)


class Text(SVGItem):
  def __init__(self, report, text, x, y):
    super(Text, self).__init__(report, text)
    self.set_attrs({"x": x, "y": y, 'fill': 'black'})
    self.html_objs = []

  def line(self, text, x, y):
    """
    Description:
    ------------
    The SVG <tspan> element defines a subtext within a <text> element or another <tspan> element.
    It allows for adjustment of the style and/or position of that subtext as needed.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG/Element/tspan

    Attributes:
    ----------
    :param text: String. The text to be added to the container
    :param x: Float. The x coordinate of the starting point of the text baseline.
    :param y: Float. The y coordinate of the starting point of the text baseline.

    :rtype: TSpan
    """
    self.html_objs.append(TSpan(self._report, text, x, y))
    return self.html_objs[-1]

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<text %s>%s%s</text>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val, str_c)


class TSpan(SVGItem):
  def __init__(self, report, text, x, y):
    super(TSpan, self).__init__(report, text)
    self.set_attrs({"x": x, "y": y, 'fill': 'black'})
    self.html_objs = []

  def __str__(self):
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<tspan %s>%s%s</tspan>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val, str_c)


class Path(SVGItem):
  def __init__(self, report, x, y, fill, origin, bespoke_path):
    super(Path, self).__init__(report, "")
    self.set_attrs({'fill': fill, "stroke": report.theme.greys[-1], "stroke-width": 1})
    if bespoke_path:
      if type(bespoke_path) != list:
        bespoke_path = [bespoke_path]
      self.__path = bespoke_path
    else:
      self.__path = ["M%s %s" % (x, y)]
    self.html_objs = []
    self.origin = origin

  def markers(self, marker_code):
    """

    :param marker_code:
    """
    self.set_attrs(name="marker-start", value=marker_code)
    self.set_attrs(name="marker-mid", value=marker_code)
    self.set_attrs(name="marker-end", value=marker_code)
    return self

  def line_to(self, x, y):
    if self.origin is not None:
      x = self.origin[0] + x
      y = self.origin[1] - y
    self.__path.append("L%s %s" % (x, y))
    return self

  def horizontal_line_to(self, x):
    if self.origin is not None:
      x = self.origin[0] + x
    self.__path.append("H%s" % x)
    return self

  def vertical_line_to(self, y):
    if self.origin is not None:
      y = self.origin[1] - y
    self.__path.append("V%s" % y)
    return self

  def move_to(self, x, y):
    if self.origin is not None:
      x = self.origin[0] + x
      y = self.origin[1] - y
    self.__path.append("M%s %s" % (x, y))
    return self

  def curve_to(self, x1, y1, x2, y2):
    """

    https://www.w3.org/TR/svg-paths/

    :param x1:
    :param y1:
    :param x2:
    :param y2:
    :return:
    """
    if self.origin is not None:
      x1 = self.origin[0] + x1
      y1 = self.origin[1] - y1
      x2 = self.origin[0] + x2
      y2 = self.origin[1] - y2
    self.__path.append("C%s %s %s %s" % (x1, y1, x2, y2))
    return self

  def smooth_curve_to(self, x1, y1, x2, y2):
    """

    https://www.w3.org/TR/svg-paths/

    :param x1:
    :param y1:
    :param x2:
    :param y2:
    :return:
    """
    if self.origin is not None:
      x1 = self.origin[0] + x1
      y1 = self.origin[1] - y1
      x2 = self.origin[0] + x2
      y2 = self.origin[1] - y2
    self.__path.append("S%s %s %s %s" % (x1, y1, x2, y2))
    return self

  def quadratic_bezier_curve_to(self, x1, y1, x2, y2):
    """

    https://www.w3.org/TR/svg-paths/

    :param x1:
    :param y1:
    :param x2:
    :param y2:
    :return:
    """
    if self.origin is not None:
      x1 = self.origin[0] + x1
      y1 = self.origin[1] - y1
      x2 = self.origin[0] + x2
      y2 = self.origin[1] - y2
    self.__path.append("Q%s %s %s %s" % (x1, y1, x2, y2))
    return self

  def smooth_quadratic_bezier_curve_to(self, x, y, absolute=True):
    """

    https://www.w3.org/TR/svg-paths/

    :param x:
    :param y:
    :param absolute:
    :return:
    """
    if self.origin is not None:
      x = self.origin[0] + x
      y = self.origin[1] - y
    self.__path.append("T%s %s" % (x, y))
    return self

  def close_path(self):
    self.__path.append("Z")
    return self

  def __str__(self):
    self.set_attrs(name="d", value="".join(self.__path))
    str_c = "".join([h.html() if hasattr(h, 'html') else str(h) for h in self.html_objs])
    return "<path %s>%s%s</path>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val, str_c)


class AnimateTransform(Html.Html):

  def __init__(self, report, attributeName, type,  from_pos, to_pos, duration, repeatCount):
    super(AnimateTransform, self).__init__(report, "")
    self.set_attrs({"attributeName": attributeName, "type": type, "from": from_pos, "to": to_pos,
                    "dur": "%ss" % duration, "repeatCount": repeatCount})

  def __str__(self):
    return "<animateTransform %s />" % self.get_attrs(pyClassNames=self.style.get_classes())


# https://stackoverflow.com/questions/6725288/svg-text-inside-rect


class Animate(Html.Html):
  # https://css-tricks.com/guide-svg-animations-smil/

  def __init__(self, report, attributeName, type,  from_pos, to_pos, duration, repeatCount):
    super(Animate, self).__init__(report, "")
    self.set_attrs({"attributeName": attributeName, "type": type, "from": from_pos, "to": to_pos,
                    "dur": "%ss" % duration, "repeatCount": repeatCount})

  def __str__(self):
    return "<animateTransform %s />" % self.get_attrs(pyClassNames=self.style.get_classes())



"""
CSS Style module for the Icons components
"""

from epyk.core.css.styles.classes import CssStyle

from epyk.core.css import Defaults as Defaults_css


class CssIcon(CssStyle.Style):
  _attrs = {'margin': '4px 10px', 'cursor': 'pointer'}

  def customize(self):
    self.hover.css({"color": self.rptObj.theme.colors[5]})


class CssStdIcon(CssStyle.Style):
  _attrs = {'display': 'inline-block', 'margin': '0 0 0 20px', 'cursor': 'pointer'}

  def customize(self):
    self.css({"color": self.rptObj.theme.colors[5],
              'font-size': '%s%s' % (Defaults_css.Icon.big, Defaults_css.Icon.unit)})
    self.hover.css({"color": self.rptObj.theme.colors[5]})


class CssSmallIcon(CssStyle.Style):
  _attrs = {'display': 'inline-block', 'margin': '0 5px 0 0', 'cursor': 'pointer'}

  def customize(self):
    self.hover.css({"color": self.rptObj.theme.colors[5],
                    'font-size': '%s%s' % (Defaults_css.Icon.small, Defaults_css.Icon.unit)})


class CssSmallIconRight(CssStyle.Style):
  _attrs = {'display': 'inline-block', 'margin': '0 0 0 5px', 'cursor': 'pointer', 'float': 'right'}

  def customize(self):
    self.hover.css({"color": self.rptObj.theme.colors[5],
                    'font-size': '%s%s' % (Defaults_css.Icon.small, Defaults_css.Icon.unit)})


class CssSmallIconRed(CssStyle.Style):
  attrs = {'display': 'inline-block', 'margin': '0 5px 0 0', 'cursor': 'pointer'}

  def customize(self):
    self.css({"color": self.rptObj.theme.danger[1],
              'font-size': '%s%s' % (Defaults_css.Icon.small, Defaults_css.Icon.unit)})
    self.hover.css({"color": self.rptObj.theme.danger[1]})


class CssOutIcon(CssStyle.Style):
  attrs = {'display': 'inline-block', 'margin': '0 0 0 20px', 'cursor': 'pointer'}

  def customize(self):
    self.css({"color": self.rptObj.theme.danger[1],
              'font-size': '%s%s' % (Defaults_css.Icon.normal, Defaults_css.Icon.unit)})
    self.hover.css({"color": self.rptObj.theme.danger[1]})


class CssBigIcon(CssStyle.Style):
  attrs = {'display': 'inline-block', 'margin': '0 10px 0 10px', 'cursor': 'pointer'}

  def customize(self):
    self.css({"color": self.rptObj.theme.danger[1],
              'font-size': '%s%s' % (Defaults_css.Icon.big, Defaults_css.Icon.unit)})
    self.hover.css({"color": self.rptObj.theme.danger[1]})




from epyk.core.css import Properties
from epyk.core.css import Defaults_css


class Attrs(Properties.CssMixin):
  def __init__(self, htmlObj):
    self.htmlObj, self.attrs = self, {}
    self.orign_htmlObj = htmlObj

  def css(self, attrs):
    """

    :param attrs:
    :return:
    """
    if not isinstance(attrs, dict):
      return self.attrs.get(attrs)

    for k, v in attrs.items():
      self.attrs[k] = v

  def __str__(self):
    """

    """
    css_tag = []
    for k, v in self.attrs.items():
      css_tag.append("%s:%s" % (k, v))
    return ";".join(css_tag)


class Commons(Attrs):

  def __init__(self, htmlObj):
    super(Commons, self).__init__(htmlObj)
    self.font_size = 'inherit'
    self.font_family = 'inherit'
    self.box_sizing = 'border-box'


class Empty(Attrs):

  def __init__(self, htmlObj):
    super(Empty, self).__init__(htmlObj)


class Body(Attrs):

  def __init__(self, htmlObj):
    super(Body, self).__init__(htmlObj)
    self.font_size = Defaults_css.font()
    self.font_family = Defaults_css.Font.family
    self.margin = 0





"""
Module in charge of monitoring the internal data sources in the framework.

This module will mainly monitor
  - The temp files: Cached files
  - The downloaded or saved files

For the cached the format will be using object whereas files will be stored without
any specific conversion. The module in charge of reading the file will be in
charge of parsing the data.

"""

import os
import sys
import json
import pickle
import importlib

from epyk.core.data import DataDb
from epyk.core.data import DataGrpc
from epyk.core.data import DataOffice
from epyk.core.js.Imports import requires
from epyk.core.js.packages import JsQuery


class DataSrc(object):
  class __internal(object):
    _props = {}

  def __init__(self, report=None):
    self._report = report if report is not None else self.__internal()

  @property
  def db(self):
    return DataDb.DataDb(self._report)

  @property
  def office(self):
    return DataOffice.DataOffice(self._report)

  def from_cache(self, code, is_secured=False, report_name=None):
    """
    Loads data from a cached files

    :param code: The code for the data
    :param is_secured: Optional, boolean to set if the file should be secured. Default False
    :param report_name: Optional, the environment in which cache are stored. Default current one

    :return: Return the data
    """
    if getattr(self._report, "run", None) is not None:
      report_name = report_name or self._report.run.report_name
      path = self._report.run.local_path
      if report_name != self._report.run.report_name:
        path = self._report.run.local_path.replace(self._report.run.report_name, report_name)
      cache_path = os.path.join(path, "tmp")
      if not os.path.exists(cache_path):
        os.mkdir(cache_path) # Create the path to store the temp files
      file_path = os.path.join(path, "tmp", code)
      if os.path.exists(file_path):
        file_obj = open(file_path, 'rb')
        return pickle.load(file_obj)

      return None

  def save_cache(self, data, code, is_secured=False, if_missing=True):
    """
    Temporary files are saved in a pickle manner in order to avoid having to parse those files again.

    :param data: The data to be saved
    :param code: The code for the data
    :param is_secured: Optional, boolean to set if the file should be secured. Default False
    :param if_missing: Optional, boolean to set the fact that caches are only saved if missing
    """
    if getattr(self._report, "run", None) is not None:
      cache_path = os.path.join(self._report.run.local_path, "tmp")
      if not os.path.exists(cache_path):
        os.mkdir(cache_path)  # Create the path to store the temp files
      if if_missing and not os.path.exists(os.path.join(cache_path, code)):
        file_obj = open(os.path.join(cache_path, code), 'wb')
        pickle.dump(data, file_obj)

  def from_file(self, filename, isSecured=False, report_name=None):
    """
    Return the file

    :param filename: The filename
    :param isSecured: Optional, Check if the file is secured or not
    :param report_name: The environment with the file
    :return: The file object
    """
    if getattr(self._report, "run", None) is not None:
      report_name = report_name or self._report.run.report_name
      path = self._report.run.local_path
      if report_name != self._report.run.report_name:
        path = self._report.run.local_path.replace(self._report.run.report_name, report_name)
      cachePath = os.path.join(path, "data")
      if not os.path.exists(cachePath):
        os.mkdir(cachePath) # Create the path to store the temp files
      filePath = os.path.join(path, "data", "%s.csv" % filename)
      return open(filePath)

  def from_source(self, http_data, fileName, fncName="getData", report_name=None, folder="sources", path=None):
    """
    Returns data from a internal data service defined in the sources folder

    :param http_data: The input data for the service
    :param fileName: The service file name
    :param fncName: Optional, the function name in the service. Default getData
    :param report_name: Optional, the report name. Default the current one
    :param folder: Optional, the folder with the services. Default sources
    :param path: Optional, the path to be added to the python system path
    :return: The data
    """
    fileName = fileName.replace(".py", "")
    if path is not None:
      if path not in sys.path:
        sys.path.append(path)
      if folder is not None:
        mod = importlib.import_module("%s.%s" % (folder, fileName))
      else:
        mod = importlib.import_module(fileName)
    else:
      if report_name is None:
        report_name = self._report.run.report_name
      mod = importlib.import_module("%s.%s.%s" % (report_name, folder, fileName))
    return getattr(mod, fncName)(self._report, http_data)

  def from_post_source(self, script, data=None, successFncs=None, udpate_freq=None, interval_name=None):
    """

    :param script:
    :param data:
    :param successFncs:
    :param udpate_freq: Optional, Set the data update frequency in second
    :return:
    """
    if udpate_freq is not None:
      return self._report.js.window.setInterval(JsQuery.JQuery(self._report).getPyScript(script, data, successFncs=successFncs), milliseconds=udpate_freq * 1000).setVar(interval_name)

    return JsQuery.JQuery(self._report).getPyScript(script, data, successFncs=successFncs)

  def from_get(self, url, data=None, code=None):
    """

    """
    return JsQuery.JQuery(self._report).get(url, data)

  def pdf(self, filename, path=None):
    """
    Read a pdf file

    This will require an external module PyPDF2
    Example
    data = rptObj.data.pdf("document.pdf", r"")
    data.getPage(0)

    Related Pages:

			https://www.geeksforgeeks.org/working-with-pdf-files-in-python/

    :param filename: The pdf file name
    :param path: The file path
    :return: A pdf object from PyPDF2
    """
    pyPDF2 = requires("PyPDF2", reason='Missing Package', install='PyPDF2', sourceScript=__file__, raiseExcept=True)
    pdf_data = pyPDF2.PdfFileReader(os.path.join(path, filename))
    return pdf_data

  # --------------------------------------------------------------------------------
  # generic system entry point
  #

  def soap(self, wsdl):
    """
    Interface to a SOAP server.

    This function will require an external python package zeep to use SOAP

    Example
    soap = rptObj.data.soap("http://www.soapclient.com/xml/soapresponder.wsdl")
    soap.Method1('Zeep', 'is cool')

    Related Pages:

			https://en.wikipedia.org/wiki/SOAP
    https://python-zeep.readthedocs.io/en/master/

    :param wsdl: The wsdl service url
    :rtype: zeep.service
    :return: The SOAP services
    """
    soap = requires("zeep", reason='Missing Package', install="zeep", sourceScript=__file__, raiseExcept=True)
    return soap.Client(wsdl).service

  def rest(self, url, data=None, method=None, encoding='utf-8', headers=None, unverifiable=False, proxy=None):
    """
    Interface to a REST server.

    Test with a online server can be done here https://jsonplaceholder.typicode.com/

    Example
    rptObj.data.rest("https://jsonplaceholder.typicode.com/posts/1", method="PUT")

    Related Pages:

			https://jsonrpcclient.readthedocs.io/en/latest/api.html

    :param url: The REST service url
    :param data: The input data for the service
    :return:
    """
    return json.loads(self._report.py.request(url, data, method, encoding, headers, unverifiable, proxy=proxy))

  def socket(self, data, host='localhost', port=5000, encoding='utf-8'):
    """

    :param data: The input data for the service
    :param host: The service host name (e.g localhost)
    :param port: The service port
    :param encoding:
    :return:
    """
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    s.sendall(json.dumps(data).encode(encoding=encoding))
    result = s.recv(1024)
    s.close()
    return result



  def websocket(self):
    pass

  def rss(self, url, proxy=None, method="GET"):
    """
    Entry point to retrieve RSS feeds.

    This module will require beautifulsoup4 as external package

    Example
    xml_soup = rptObj.data.rss("http://feeds.reuters.com/reuters/businessNews")
    for title in xml_soup.findAll('title'):
      print(title)

    Related Pages:

			https://pypi.org/project/beautifulsoup4/

    :param url: The url of the html page
    :param method: Optional, The request method. Default method GET
    :return: A xml object
    """
    bs4 = requires("bs4", reason='Missing Package', install='beautifulsoup4', sourceScript=__file__, raiseExcept=True)
    headers = {'User-Agent': 'Mozilla/5.0', 'accept': 'application/xml;q=0.9, */*;q=0.8'}
    response = self._report.py.requests.get(url, headers=headers, proxy=proxy)
    xml_soup = bs4.BeautifulSoup(response,)
    return xml_soup

  def webscrapping(self, url, parser="html.parser", proxy=None, method=None):
    """
    Entry point to retrieve data from any website.

    This module will require beautifulsoup4 as external package

    Example
    rptObj.data.webscrapping("https://www.w3schools.com/colors/default.asp")
    xml_soup.findAll('title')

    Related Pages:

			https://pypi.org/project/beautifulsoup4/

    :param url: The url of the html page
    :param parser: The output data parser
    :return: A xml object
    """
    bs4 = requires("bs4", reason='Missing Package', install='beautifulsoup4', sourceScript=__file__, raiseExcept=True)
    headers = {
      'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
      'Accept-Encoding': 'none',
      'Accept-Language': 'en-US,en;q=0.8',
      'Connection': 'keep-alive',
      'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.3',
      'User-Agent': 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.3'}
    response = self._report.py.request(url, headers=headers, method=method, proxy=proxy)
    xml_soup = bs4.BeautifulSoup(response, parser)
    return xml_soup



  def rpc(self, url, data=None, headers=None, is_secured=False):
    """
    Interface to a RPC server.

    This is using the external python package jsonrpcclient (https://jsonrpcclient.readthedocs.io/en/latest/)

    Example

    Related Pages:

			https://en.wikipedia.org/wiki/Rapid_control_prototyping
    https://gurujsonrpc.appspot.com/
    https://jsonrpcclient.readthedocs.io/en/latest/

    :param url: The RPC service url
    :param data: The input data for the service
    :return:
    """
    http_client = requires("jsonrpcclient.clients.http_client", reason='Missing Package', install="jsonrpcclient[requests]", sourceScript=__file__, raiseExcept=True)
    client = http_client.HTTPClient(url)
    if headers is not None:
      client.session.headers.update(headers)
    if is_secured:
      client.session.auth = ("user", "pass")
    if data is None or "method" not in data:
      raise Exception("data must of a method defined")

    return client.send(json.dumps(data))

  def grpc(self, serviceName, path, module, host="localhost", port=50051):
    """
    Interface to a GRPC server.

    Example
    grpc = rptObj.data.grpc(serviceName="GreeterStub", module="helloworld_pb2_grpc", path="")
    data = grpc.imp("helloworld_pb2").HelloRequest(name="Test")
    print(grpc.request("SayHello", data))

    Related Pages:

			https://grpc.io/docs/tutorials/basic/python/
    https://grpc.io/docs/quickstart/python.html

    :param serviceName: The Service name (the class name in the python module)
    :param path: The path with the GRPC features
    :param module: The python module name for the service
    :param host: The service host name (e.g localhost)
    :param port: The service port
    :return: A GRPC wrapped object
    """
    requires("grpc", reason='Missing Package', install='grpcio', sourceScript=__file__, raiseExcept=True)
    return DataGrpc.DataGrpc(serviceName, path, module, host, port)




from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesSelect


class CatalogSelect(Catalog.CatalogGroup):
  def base(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectStyle)

  def button(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectButton)

  def toggle(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectToggle)

  def search_box_input(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectSearchBoxInput)

  def option(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectOption)

  def item(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectOptionItems)

  def active(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectOptionActive)

  def selected(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectOptionSelected)

  def option_filter(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectFilterOption)

  def outline(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectOutline)

  def status(self):
    """  """
    return self._set_class(CssStylesSelect.CssSelectStatus)

